#   Simple Backup - provides access to TAR functionality
#
#   Copyright (c)2008-2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2008: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`tar` --- provides access to TAR functionality
===================================================

.. module:: tar
   :synopsis: provides access to TAR functionality
.. moduleauthor:: Ouattara Oumar Aziz (alias wattazoum) <wattazoum@gmail.com>
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>

"""

from gettext import gettext as _
import time
import os
import subprocess
import tempfile
import shutil
import re

from datetime import datetime


from sbackup.core.ConfigManager import ConfigurationFileHandler

from sbackup.fs_backend import fam

from sbackup.util.log import LogFactory
from sbackup.util.structs import SBdict
from sbackup.util import exceptions
from sbackup.util.exceptions import SBException
from sbackup.util import constants
from sbackup.util import system
from sbackup import util
from sbackup.util import local_file_utils
from sbackup.util import structs
from sbackup.util import log


GZIP_COMPRESSION_SPEED = "-1"


_FOP = fam.get_file_operations_facade_instance()


def getArchiveType(archive):
    """Determines the type of an archive by its file extension.
     
    @param archive: Full path to file to check  
    @return: tar, gzip, bzip2 or None
    @rtype: String
    """
    _res = None
    if archive.endswith(".bz2"):
        _res = "bzip2"
    elif archive.endswith(".gz") or archive.endswith(".tgz"):
        _res = "gzip"
    elif archive.endswith(".tar"):
        _res = "tar"
    return _res

def extract(sourcear, eff_local_sourcear, restore_file, dest , bckupsuffix = None, splitsize = None):
    """Extract from source archive the file "file" to dest.
    
    @param sourcear: path of archive
    @param file:
    @param dest:
    @param bckupsuffix: If set a backup suffix option is set to backup existing files
    @param splitsize: If set the split options are added supposing the size of the archives is this variable
    @type splitsize: Integer in 1024 Bytes
    """
    _logger = log.LogFactory().getLogger()
    _logger.debug("input param `sourcear`: %s" % sourcear)
    _logger.debug("input param `eff_local_sourcear`: %s" % eff_local_sourcear)
    restore_file = restore_file.lstrip(_FOP.pathsep)    # strip leading separator (as TAR did before)
    # tar option  -p, --same-permissions, --preserve-permissions:
    # ignore umask when extracting files (the default for root)

    options = ["-xp", "--ignore-failed-read", "--backup=existing", "--totals"]

    archType = getArchiveType(sourcear)
    _logger.debug("Archive type: %s" % archType)
    if archType == "tar" :
        pass
    elif archType == "gzip" :
        options.insert(1, "--gzip")
    elif archType == "bzip2" :
        options.insert(1, "--bzip2")
    else :
        raise SBException (_("Invalid archive type."))

    if system.is_superuser():
        options.append("--same-owner")
    if dest :
        options.append("--directory=" + dest)
    else :
        options.append("--directory=" + _FOP.pathsep)
    if bckupsuffix :
        options.append("--suffix=" + bckupsuffix)

    if splitsize > 0:
        options.extend(["-L %s" % splitsize , "-F %s" % util.get_resource_file("multipleTarScript")])
        if eff_local_sourcear is None:
            raise exceptions.FileAccessException(_("Effective path for `%s` is not available") % sourcear)

    _launcher = TarBackendLauncherSingleton()
    if eff_local_sourcear is None:
        _launcher.set_stdin_file(sourcear)
    else:
        options.append("--file=%s" % eff_local_sourcear)

    options.append(restore_file)

    _launcher.launch_sync(options, env = {})
    retVal = _launcher.get_returncode()
    outstr = _launcher.get_stdout()
    errStr = _launcher.get_stderr()

    __finish_tar(retVal, outstr, errStr)


# extract2 is currently (series 0.2) only used in SnapshotManager.makeTmpTAR
def extract2(sourcear, fileslist, dest, bckupsuffix = None, additionalOpts = None):
    """Extract the files listed in the 'fileslist' file to dest. This method 
    has been created to optimize the time spent by giving to tar a complete 
    list of file to extract.
    
    Use this if ever you have to extract more than 1 dir .
    
    @param sourcear:
    @param fileslist: a path to the file containing the list
    @param dest: destination
    @param bckupsuffix: 
    @param additionalOpts: a list of options to add
    """
    # tar option  -p, --same-permissions, --preserve-permissions:
    # ignore umask when extracting files (the default for root)
    options = ["-xp", "--ignore-failed-read", '--backup=existing']

    archType = getArchiveType(sourcear)
    if archType == "tar" :
        pass
    elif archType == "gzip" :
        options.insert(1, "--gzip")
    elif archType == "bzip2" :
        options.insert(1, "--bzip2")
    else :
        raise SBException (_("Invalid archive type."))

    if system.is_superuser():
        options.append("--same-owner")

    if dest:
        options.append("--directory=" + dest)
    else:
        options.append("--directory=" + _FOP.pathsep)

    if bckupsuffix:
        options.append("--suffix=" + bckupsuffix)

    if additionalOpts and type(additionalOpts) == list:
        options.extend(additionalOpts)

    options.extend(['--file=' + sourcear, '--null',
                    '--files-from=' + _FOP.normpath(fileslist)])

    outStr, errStr, retval = util.launch("tar", options)
    if retval != 0 :
        LogFactory.getLogger().debug("output was : " + outStr)
        raise SBException("Error when extracting : " + errStr)
    if outStr.strip() == "":
        outStr = "(nothing)"
    LogFactory.getLogger().debug("output was: " + outStr)


def appendToTarFile(desttar, fileslist, workingdir, additionalOpts):
    """
    @param desttar: The tar file to wich append
    @param fileslist The files to append, can be a list of files too
    @type fileslist: list
    @param workingdir: the dir to move in before appending the dir ( usefun for relative paths)
    @param additionalOpts: a list of additional option to append (will be append before changing the working dir)
    """
    options = ["--append", "--ignore-failed-read"]

    archType = getArchiveType(desttar)
    if archType == "tar" :
        pass
    elif archType == "gzip" :
        options.insert(1, "--gzip")
    elif archType == "bzip2" :
        options.insert(1, "--bzip2")
    else :
        raise SBException (_("Invalid archive type."))

    if additionalOpts and type(additionalOpts) == list :
        options.extend(additionalOpts)

    options.extend(['--file=' + desttar, '--null'])

    if workingdir:
        options.append("--directory=" + workingdir)

    options.append('--files-from=' + _FOP.normpath(fileslist))

    outStr, errStr, retval = util.launch("tar", options)
    if retval != 0 :
        LogFactory.getLogger().debug("output was : " + outStr)
        raise SBException("Error when extracting : " + errStr)
    if outStr.strip() == "":
        outStr = "(nothing)"
    LogFactory.getLogger().debug("output was: " + outStr)


def __prepare_common_opts(snapshot, targethandler, publish_progress, use_io_pipe):
    """Prepares common TAR options used when full or incremental
    backups are being made.  
    
    :param snapshot: The snapshot to fill in
    :return: a list of options to be use to launch tar
    
    :todo: Check whether it's necessary to escape white spaces in path names!
    
    """
    # don't escape spaces i.e. do not replace them with '\ '; this will fail
    # take care where to insert additional options (e.g. --gzip)

    tdir = snapshot.getPath()
    _snpname = snapshot.getName()
    options = ["-cS", "--directory=%s" % _FOP.pathsep,
               "--ignore-failed-read",
               "--no-check-device",
               "--blocking-factor", str(constants.TAR_BLOCKING_FACTOR),
               "--totals"
              ]

    # includes and excludes
    tmp_incl = _FOP.normpath(ConfigurationFileHandler().get_user_tempdir(),
                                 _FOP.get_basename(snapshot.getIncludeFListFile()))
    tmp_excl = _FOP.normpath(ConfigurationFileHandler().get_user_tempdir(),
                                 _FOP.get_basename(snapshot.getExcludeFListFile()))

    LogFactory.getLogger().debug("Temporary includes file: `%s`" % tmp_incl)
    LogFactory.getLogger().debug("Temporary excludes file: `%s`" % tmp_excl)

    if not _FOP.path_exists(tmp_incl):
        raise SBException("Temporary includes.list does not exist")
    if not _FOP.path_exists(tmp_excl):
        raise SBException("Temporary excludes.list does not exist")

    options.append('--files-from=%s' % tmp_incl)
    options.append('--exclude-from=%s' % tmp_excl)

    if snapshot.isFollowLinks() :
        options.append("--dereference")

    archivename = "files.tar"
    if snapshot.getFormat() == "gzip":
        options.insert(0, "--gzip")
        archivename += ".gz"
    elif snapshot.getFormat() == "bzip2":
        options.insert(0, "--bzip2")
        archivename += ".bz2"
    elif snapshot.getFormat() == "none":
        pass
    else:
        LogFactory.getLogger().debug("Setting compression to default 'none'")

    _ar_path = _FOP.normpath(tdir, archivename)
    if use_io_pipe:
        pass
    # write to socket file object?
    else:
        _larpath = targethandler.get_eff_fullpath(_snpname, archivename)
        if _larpath is None:
            raise exceptions.FileAccessException(_("Unable to get effective path for `%s`") % _ar_path)
        options.append('--file=%s' % _larpath)

    # progress signal
#TODO: improve calculation of number of checkpoints based on estimated transfer rate.
    if publish_progress is True:
        one_mb = 1000000
        one_gb = 1000 * one_mb
        _snpsize = snapshot.get_space_required()
        if _snpsize < (50 * one_mb):
            _ncheckp = 10
        elif _snpsize < (500 * one_mb):
            _ncheckp = 50
        elif _snpsize < (5 * one_gb):
            _ncheckp = 250
        else:
            _ncheckp = 500

        _checkpsize = _snpsize / _ncheckp
        _checkp = int(_checkpsize / constants.TAR_RECORDSIZE)
        if _checkp > 0:
            # LP: #875634
            # Sometimes the calculation of the space required goes badly wrong
            # and we could end-up with _checkp=1 for 30GB to save, making tar
            # VERY slow. Fix: set minimum value to 100
            _checkp = max(100, _checkp)
            options.append("--checkpoint=%s" % _checkp)
            _progessf = util.get_resource_file(resource_name = "sbackup-progress")
            options.append('--checkpoint-action=exec=%s' % _progessf)

    if LogFactory.getLogger().isEnabledFor(5):
        options.append("--verbose")

    return (options, _ar_path, tmp_incl, tmp_excl)


def __add_split_opts(snapshot, options, size):
    """
    Compiles and add the split management options to the TAR line.
    Valid for read and create actions
    @param snapshot: The snapshot to process
    @type snapshot: Snapshot
    @param options: the option in which to append
    @type options: list
    @param size: the size of each part (in KiB)
    @type size: int
    @raise SBException: if the snapshot format is other than none
    """
    if snapshot.getFormat() != "none" :
        raise SBException(_("For the moment split functionality is not compatible with compress option."))
    options.extend(["-L %s" % str(size), "-F %s" % util.get_resource_file("multipleTarScript")])
    return options


def __copy_temp_snarfile_into_snapshot(tmp_snarfile, snarfile):
    try:
        _FOP.copyfile(tmp_snarfile, snarfile)
    except exceptions.CopyFileAttributesError:
        LogFactory.getLogger().warning(_("Unable to change permissions for file '%s'.")\
                                    % snarfile)


def __remove_tempfiles(tmp_files):
    for _tmpf in tmp_files:
        try:
            _FOP.delete(_tmpf)
        except (OSError, IOError), error:
            LogFactory.getLogger().warning(_("Unable to remove temporary file `%(file)s`: %(error)s")\
                                            % { 'file' : _tmpf, 'error': error })


def mk_archive(snapshot, targethandler, publish_progress, supports_publish):
    if snapshot.isfull():
        _mk_tar_full(snapshot, targethandler, publish_progress, supports_publish)
    else:
        _mk_tar_incr(snapshot, targethandler, publish_progress, supports_publish)

def _mk_tar_incr(snapshot, targethandler, publish_progress, supports_publish):
    """
    Launch a TAR incremental backup
    @param snapshot: the snapshot in which to make the backup
    @raise SBException: if there was a problem with tar
    """
    LogFactory.getLogger().info(_("Launching TAR to make incremental backup."))

    # LP #638072: no wildcards in excludes list
    _env = { "GZIP" : GZIP_COMPRESSION_SPEED,
             "TAR_OPTIONS" : "--no-wildcards --anchored --no-wildcards-match-slash" }
    _use_io_pipe = targethandler.get_use_iopipe()
    _splitsize = snapshot.getSplitedSize()
    if _splitsize > 0:
        _use_io_pipe = False
        publish_progress = publish_progress and supports_publish
        _env["SBACKUP_VOLUME_SIZE"] = str(_splitsize * constants.TAR_VOLUME_SIZE_UNIT_IN_BYTES)

    options, ar_path, tmp_incl, tmp_excl = __prepare_common_opts(snapshot, targethandler,
                                                                 publish_progress, _use_io_pipe)

    if _splitsize > 0:
        options = __add_split_opts(snapshot, options, _splitsize)

    base_snarfile = snapshot.getBaseSnapshot().getSnarFile()
    snarfile = snapshot.getSnarFile()
    tmp_snarfile = _FOP.normpath(ConfigurationFileHandler().get_user_tempdir(),
                                 _FOP.get_basename(snarfile))

    LogFactory.getLogger().debug("Snapshot's base snarfile: %s" % base_snarfile)
    LogFactory.getLogger().debug("Snapshot's snarfile: %s" % snarfile)
    LogFactory.getLogger().debug("Temporary snarfile: %s" % tmp_snarfile)

    # For an INC backup the base SNAR file should exists
    if not _FOP.path_exists(base_snarfile) :
        LogFactory.getLogger().error(_("Unable to find the SNAR file to make an incremental backup."))
        LogFactory.getLogger().error(_("Falling back to full backup."))
        _mk_tar_full(snapshot, targethandler, publish_progress, _use_io_pipe)
    else:
        try:
            _FOP.copyfile(base_snarfile, tmp_snarfile)
        except exceptions.CopyFileAttributesError:
            LogFactory.getLogger().warning(_("Unable to change permissions for file '%s'.")\
                                    % tmp_snarfile)

        # check (and set) the permission bits; necessary if the file's origin
        # does not support user rights (e.g. some FTP servers, file systems...)
#FIXME: use generalized File operations here!
        if not local_file_utils.path_writeable(tmp_snarfile):
            local_file_utils.chmod(tmp_snarfile, 0644)
        # create the snarfile within a local directory; necessary if the
        # backup target does not support 'open' within the TAR process and
        # would fail
        options.append("--listed-incremental=" + tmp_snarfile)

        # launch TAR with empty environment
        _launcher = TarBackendLauncherSingleton()
        if _use_io_pipe:
            _launcher.set_stdout_file(ar_path)
        try:
            _launcher.launch_sync(options, env = _env)
            retVal = _launcher.get_returncode()
            outstr = _launcher.get_stdout()
            errStr = _launcher.get_stderr()
            __finish_tar(retVal, outstr, errStr)
            __copy_temp_snarfile_into_snapshot(tmp_snarfile, snarfile)

        finally:
            __remove_tempfiles([tmp_snarfile, tmp_incl, tmp_excl])


def _mk_tar_full(snapshot, targethandler, publish_progress, supports_publish):
    """Convenience function that launches TAR to create a full backup.

    @param snapshot: the snapshot in which to make the backup

    @raise SBException: if there was a problem with tar
    """
    LogFactory.getLogger().info(_("Launching TAR to make a full backup."))

    # LP #638072: no wildcards in excludes list
    _env = { "GZIP" : GZIP_COMPRESSION_SPEED,
             "TAR_OPTIONS" : "--no-wildcards --anchored --no-wildcards-match-slash" }
    _use_io_pipe = targethandler.get_use_iopipe()
    _splitsize = snapshot.getSplitedSize()
    if _splitsize > 0:
        _use_io_pipe = False
        publish_progress = publish_progress and supports_publish
        _env["SBACKUP_VOLUME_SIZE"] = str(_splitsize * constants.TAR_VOLUME_SIZE_UNIT_IN_BYTES)

    options, ar_path, tmp_incl, tmp_excl = __prepare_common_opts(snapshot, targethandler,
                                                                 publish_progress, _use_io_pipe)

    if _splitsize > 0:
        options = __add_split_opts(snapshot, options, _splitsize)

    snarfile = snapshot.getSnarFile()
    tmp_snarfile = _FOP.normpath(ConfigurationFileHandler().get_user_tempdir(),
                                 _FOP.get_basename(snarfile))

    LogFactory.getLogger().debug("Snapshot's snarfile: %s" % snarfile)
    LogFactory.getLogger().debug("Temporary snarfile: %s" % tmp_snarfile)

    # For a full backup the SNAR file shouldn't exists
    if _FOP.path_exists(snarfile) :
        _FOP.delete(snarfile)
    if _FOP.path_exists(tmp_snarfile) :
        _FOP.delete(tmp_snarfile)

    options.append("--listed-incremental=%s" % tmp_snarfile)

    # launch TAR with empty environment
    _launcher = TarBackendLauncherSingleton()
    if _use_io_pipe:
        _launcher.set_stdout_file(ar_path)
    try:
        _launcher.launch_sync(options, env = _env)
        retVal = _launcher.get_returncode()
        outstr = _launcher.get_stdout()
        errStr = _launcher.get_stderr()
        __finish_tar(retVal, outstr, errStr)
        __copy_temp_snarfile_into_snapshot(tmp_snarfile, snarfile)
    finally:
        __remove_tempfiles([tmp_snarfile, tmp_incl, tmp_excl])


def __finish_tar(exitcode, output_str, error_str):
    """
    @todo: Catch failing child process (e.g. gzip) in proper way.
    """
    if output_str is None:
        output_str = ""
    if error_str is None:
        error_str = ""

    _logger = LogFactory.getLogger()
    _logger.debug("Exit code: %s" % exitcode)
    _logger.debug("Standard out: %s" % output_str)
    _logger.debug("Error out: %s" % error_str)

    _res_err = []
    if error_str != "":
        err_lst = error_str.split("\n")
        for err_item in err_lst:
            if err_item == "/bin/tar: Removing leading `/' from member names":
                _logger.info(_("Leading '/' from member names were removed."))
            elif err_item == "/bin/tar: Removing leading `/' from hard link targets":
                _logger.info(_("Leading '/' from hard link targets were removed."))
            else:
                err_item = err_item.lstrip("/bin/tar: ")
                err_item = err_item.strip()
                if err_item != "":
                    _res_err.append(err_item)

    if exitcode == 0:
        if len(_res_err) == 1:
            _logger.info(_("TAR returned a message: %s") % _res_err[0])
        elif len(_res_err) > 1:
            _logger.info(_("TAR returned messages:\n%s") % "\n".join(_res_err))
        else:
            pass
        _logger.info(_("TAR has been finished successfully."))

    elif exitcode == 1:
        if len(_res_err) == 1:
            _logger.info(_("TAR returned a warning: %s") % _res_err[0])
        elif len(_res_err) > 1:
            _logger.info(_("TAR returned warnings:\n%s") % "\n".join(_res_err))
        else:
            pass
        _logger.info(_("TAR returned warnings but has been finished successfully."))

    else:
        # list-incremental is not compatible with ignore failed read ???
        _errmsg = _("Unable to finish successfully. TAR terminated with errors.")
        _logger.error("%s\n%s(exit code: %s)" % (_errmsg, error_str, exitcode))
        raise SBException(_errmsg)


def get_dumpdir_from_list(lst_dumpdirs, filename):
    """Searchs within the given list of dumpdirs for the given filename
    and the dumpdir if found.
    
    @raise SBExcetion: if filename couldn't be found in list  
    """
#    print ">>> get_dumpdir_from_list"
#    print "  Looking for: %s" % filename
#    print "  List of dumpdirs:"
#    for _ddir in lst_dumpdirs:
#        print "    %s" % _ddir
    _res = None

    if not isinstance(lst_dumpdirs, list):
        raise TypeError("Given list of dumpdirs must be of type list. Got %s "\
                        "instead." % type(lst_dumpdirs))
    for _ddir in lst_dumpdirs:
        if not isinstance(_ddir, Dumpdir):
            raise TypeError("Element in list of dumpdirs must be of type "\
                            "Dumpdir. Got %s instead." % type(_ddir))

        if _ddir.getFilename() == filename:
            _res = _ddir
            break

    if _res is None:
        raise SBException("File '%s' was not found in given list of Dumpdirs."
                                                                    % filename)
    return _res


class TarBackendLauncherSingleton(object):
    __metaclass__ = structs.Singleton

    _cmd = "/bin/tar"

    def __init__(self):
        self._proc = None
        self._argv = []

        # files used for input/output piping
        self._stdout_f = None
        self._stdin_f = None

        # results
        self._stdout = None
        self._stderr = None
        self._returncode = None

    def set_stdin_file(self, path):
        if self._stdout_f is not None:
            raise AssertionError("Redirecting stdin and stdout is not supported")
        self._stdin_f = path

    def set_stdout_file(self, path):
        if self._stdin_f is not None:
            raise AssertionError("Redirecting stdin and stdout is not supported")
        self._stdout_f = path

    def get_pid(self):
        _pid = None
        if self._proc is not None:
            _pid = self._proc.pid
        return _pid

    def launch_sync(self, opts, env = None):
        # note: get logger instance on demand as BackendLauncher is a Singleton
        _logger = log.LogFactory.getLogger()
        if self._proc is not None:
            raise AssertionError("Another process is already running")

        self._clear_returns()
        self._argv = opts
        self._argv.insert(0, self._cmd)
        errptr, errfile = tempfile.mkstemp(prefix = "error_")

        _stdout_param = None
        _stdin_param = None

        try:
            _logger.debug("Lauching: %s" % (" ".join(self._argv)))

            if self._stdout_f is None:
                outptr, outfile = tempfile.mkstemp(prefix = "output_")
                _stdout_param = outptr
            else:
                assert self._stdin_f is None
                _logger.debug("Output archive: %s" % self._stdout_f)
                _stdout_param = subprocess.PIPE
                _ardst = _FOP.openfile_for_write(self._stdout_f)

            if self._stdin_f is not None:
                assert self._stdout_f is None
                _logger.debug("Input archive: %s" % self._stdin_f)
                _stdin_param = subprocess.PIPE
                _arsrc = _FOP.openfile_for_read(self._stdin_f)

            self._proc = subprocess.Popen(self._argv, stdin = _stdin_param, stdout = _stdout_param,
                                          stderr = errptr, env = env)
            _logger.debug("Subprocess created")
            if self._stdout_f is not None:
                _arsrc = self._proc.stdout
            if self._stdin_f is not None:
                _ardst = self._proc.stdin

            if (self._stdin_f is not None) or (self._stdout_f is not None):
                shutil.copyfileobj(_arsrc, _ardst)
                _arsrc.close()
                _ardst.close()
                self._proc.wait()
            else:
                self._proc.communicate(input = None)

        except (exceptions.BackupCanceledError, exceptions.SigTerminatedError), error:
            self.terminate()

            if (self._stdin_f is not None) or (self._stdout_f is not None):
                _arsrc.close()
                _ardst.close()
            # todo: remove code duplication
            self._returncode = self._proc.returncode
            os.close(errptr)    # Close log handle
            self._stderr = local_file_utils.readfile(errfile)
            if self._stdout_f is None:
                os.close(outptr)
                self._stdout = local_file_utils.readfile(outfile)
            local_file_utils.delete(errfile)

            self._clear_proc()
            raise error

        self._returncode = self._proc.returncode
        os.close(errptr)    # Close log handle
        self._stderr = local_file_utils.readfile(errfile)
        if self._stdout_f is None:
            os.close(outptr)
            self._stdout = local_file_utils.readfile(outfile)
        local_file_utils.delete(errfile)

        self._clear_proc()

    def _clear_returns(self):
        self._stdout = None
        self._stderr = None
        self._returncode = None

    def _clear_proc(self):
        self._proc = None
        self._stdin_f = None
        self._stdout_f = None
        self._argv = []

    def is_running(self):
        """
        :todo: Add more tests whether the process is running or not!
        """
        _res = False
        if self._proc is not None:
            # more tests here
            _res = True
        return _res

    def get_stdout(self):
        return self._stdout

    def get_stderr(self):
        return self._stderr

    def get_returncode(self):
        return self._returncode

    def terminate(self):
        if self.is_running():
            try:
                self._proc.terminate()
            except OSError, error:
                log.LogFactory.getLogger().warning(_("Unable to terminate backend process: %s") % error)


class Dumpdir(object):
    """This is actually a single dumdir entry.
    
    Here is the definition from TAR manual:
    Dumpdir is a sequence of entries of the following form: C filename \0
    where C is one of the control codes described below, filename is the name
    of the file C operates upon, and '\0' represents a nul character (ASCII 0).
    
    The white space characters were added for readability, real dumpdirs do not
    contain them. Each dumpdir ends with a single nul character.
    
    Dumpdirs stored in snapshot files contain only records of types 'Y', 'N'
    and 'D'. 

    
    @note: Is the nul character stored???
    
    @see: http://www.gnu.org/software/tar/manual/html_chapter/Tar-Internals.html#SEC173
    
    @todo: It should be distiguished between 'unchanged', 'excluded' and
           'removed'!
    @todo: Rename this class into 'DumpdirEntry'!
    @todo: Add a 'DumpdirClass' that collects entries and creates output content!
    """

    INCLUDED = 'Y'
    UNCHANGED = 'N'
    DIRECTORY = 'D'

    # The dictionary mapping control with their meanings
    __HRCtrls = {'Y':_('Included'), 'N':_('Not changed'), 'D':_('Directory')}

    def __init__(self, line):
        """
        Constructor that takes a line to create a Dumpdir.
        we will parse this line and fill the Dumpdir in
        @param line: the line (in dumpdir point of view ) to parse
        @raise Exception: when the line doesn't have the requeried format
        
        @todo: make the instance variables private!
        """
        self.control = None
        self.filename = None
        self.__set_entry(line)

    def __str__(self):
        return "%s %s" % (self.filename, self.getHumanReadableControl())

    def __repr__(self):
        return self.__str__()

    def __set_entry(self, line):
        """Private helper method that sets the given dumdir entry. It checks
        for type and length of given parameter.
        
        @todo: Add checks for validity of control etc.!
        """
        if (not isinstance(line, str)):
            raise TypeError(_("Line must be a string"))
        line = line.strip("\0")
        if len(line) < 2:
            raise ValueError("Line must contain 2 characters at minimum.")
        self.control = line[0]
        self.filename = line[1:]

    def getFilename(self):
        """
        get the filename embedded in the Dumpdir
        @return: finename
        @raise Exception: if the filename is null 
        """
        if self.filename :
            return self.filename
        else :
            raise SBException(_("Dumpdir inconsistancy : 'filename' is empty"))

    def getControl(self):
        """
        Get the control character from the DumpDir
        @return: control
        @raise Exception: if the control is null
        
        @todo: Checking against None is useless here!
        """
        if self.control:
            return self.control
        else :
            raise SBException(_("Dumpdir inconsistancy : 'control' is empty"))

    def getHumanReadableControl(self):
        """
        Get the control character as a Human readable string from the DumpDir
        """
        return self.__HRCtrls[self.getControl()]

    @classmethod
    def getHRCtrls(cls):
        """
        @return: The Human Readable control dictionary
        @rtype: dict
        
        """
        return cls.__HRCtrls


class SnapshotFile(object):
    """A snapshot file (or directory file) is created during incremental
    backups with TAR. It contains the status of the file system at the time
    of the dump and is used to determine which files were modified since the
    last backup. GNU tar version 1.20 supports three snapshot file formats.
    They are called 'format 0', 'format 1' and 'format 2'.
    
    Dumpdirs stored in snapshot files contain only records of types 'Y', 'N'
    and 'D'. 
    
    For displaying the content of an incremental backup manually use:
    tar --bzip --list --incremental --verbose --verbose --file files.tar.bz2

    @see: http://www.gnu.org/software/tar/manual/html_chapter/Tar-Internals.html#SEC177
    
    @attention: Only snapshot files in `format 2` are supported.
    
    @todo: Rename into 'DirectoryFile' or 'SnapshotDirectoryFile' since it only
           collects directories!
           
    @todo: Add class describing a SnapshotFile record!
    """
    versionRE = re.compile("GNU tar-(.+?)-([0-9]+?)")

    __SEP = '\000'
    __entrySEP = 2 * __SEP

    # Infos on indices in a record 
    REC_NFS = 0
    REC_MTIME_SEC = 1
    REC_MTIME_NANO = 2
    REC_DEV_NO = 3
    REC_INO = 4
    REC_DIRNAME = 5
    REC_CONTENT = 6


    def __init__(self, filename, writeFlag = False):
        """Constructor
         
        @param filename: the snapshot file (absolute? do we use relative paths?) file path to get the infos (SNAR file)
        @param writeFlag: if set, the file will be created in case it doesn't exist
        """
        self.header = None
        self.snpfile = None
        self.version = None

        if _FOP.path_exists(filename):
            self.snpfile = filename
        else :
            if writeFlag :
                self.snpfile = filename
                fd = _FOP.openfile_for_append(self.snpfile)
                fd.close()
            else :
                raise SBException(_("File '%s' does not exist.") % filename)

    def __str__(self):
        _str = [ "Snapshot file",
                 " Header: %s" % self.getHeader(),
#                 " Version: %s" % self.getFormatVersion(),
                 " file name: %s" % self.snpfile,
                 " Content:"
               ]
        for _item in self.parseFormat2():
            _str.append("  %s" % _item)

        return "\n".join(_str)

    def get_filename(self):
        return self.snpfile

# (actually) not used: marked for removal
#    def getFormatVersion(self):
#        """
#        Get the format version
#        @return: the version 
#        @rtype: int 
#        """
#        if self.version :
#            return self.version
#
#        self.header = self.__getHeaderInfos()
#        m = self.versionRE.match(self.header)
#        if m :
#            # we are version 1 or 2
#            # check for version 2 first
#            self.version = m.group(2)
#        else :
#            # we are version 0
#            self.version = 0
#
#        return int(self.version)
#
#    def __getHeaderInfos(self):
#        """
#        Get the first line of the snapshot file
#        @return: the first line content
#        """
#        fd = open(self.snpfile)
#        header = fd.readline()
#        fd.close()
#        return header.strip()

# not used: marked for removal
#    def parseFormat0(self):
#        """
#        Iterator method that gives each line entry
#        @warning: only compatible tar version 0 of Tar format
#        @return: [nfs,dev,inode,name]
#        """
#        fd = open(self.snpfile)
#        # skip header which is the first line in this case
#        fd.readline()
#
#        for line in fd.readlines():
#            line = line.rstrip()
#
#            nfs = line[0] # can be a + or a single white space
#
#            dev, inode, name = line[1:].split(' ', 2)
#            yield [nfs, dev, inode, name]
#
#        fd.close()

# not used: marked for removal
#    def parseFormat1(self):
#        """
#        Iterator method that gives each line entry
#        @warning: only compatible tar version 1 of Tar format
#        @return: [nfs,mtime_sec,mtime_nsec,dev,inode,name]
#        """
#        fd = open(self.snpfile)
#        # skip header which is the 2 first lines in this case
#        fd.readline()
#        fd.readline()
#
#        for line in fd.readlines() :
#            line = line.rstrip()
#
#            nfs = line[0] # can be a + or a single white space
#
#            mtime_sec, mtime_nsec, dev, inode, name = line[1:].split(' ', 4)
#            yield [nfs, mtime_sec, mtime_nsec, dev, inode, name]
#
#        fd.close()

    def parseFormat2(self):
        """Iterator method that gives each line entry in SNAR-file.
        A line contains informations about a directory and its content.
        
        @warning: only compatible tar version 2 of Tar format
        
        @return: [nfs,mtime_sec,mtime_nano,dev_no,i_no,name,contents] where contents is a list of Dumpdirs
        
        @todo: Iterator methods should have names like 'parseFormat2Iterator' or
               'parseFormat2Iter'!
        """

        def formatDumpDirs(content):
            """
            Subroutine to format a content into dump dirs
            """
            result = []
            if content :
                for d in content.rstrip('\0').split('\0'):
                    if d :
                        result.append(Dumpdir(d))
            return result

        def format(line):
            """
            subroutine to format a line including NUL char to have and array
            """
            nfs, mtime_sec, mtime_nano, dev_no, i_no, name, contents = line.lstrip("\0").split("\0", 6)
            return [nfs, mtime_sec, mtime_nano, dev_no, i_no, name, formatDumpDirs(contents)]


        fd = _FOP.openfile_for_read(self.snpfile)
        # skip header which is the first line and 2 entries separated with NULL in this case
#TODO: Handle empty files properly
        n = 0
        while n < 2 :
            c = fd.read(1)
            if len(c) != 1:
                raise SBException(_("The snarfile header is incomplete."))
            if c == '\0':
                n += 1

        currentline = ""
        c = fd.read(1)
        last_c = ''

        while c:
            currentline += c
            if c == '\0' and last_c == '\0' :
                # we got a line
                yield format(currentline)

                currentline = ''
                last_c = ''
            else :
                last_c = c
            c = fd.read(1)
        fd.close

    def getHeader(self):
        """
        Get the full header
        @return: the header (a string) or None if the file was empty
        @raise SBException: if the header is incomplete
        """
        fd = _FOP.openfile_for_read(self.snpfile)
#TODO: Handle empty files properly
        header = ""
        n = 0
        while n < 2 :
            c = fd.read(1)
            if len(c) != 1:
                raise SBException(_("The snarfile header is incomplete."))
            if c == '\0':
                n += 1
            header += c
        try:
            _FOP.close_stream(fd)
        except exceptions.FileAlreadyClosedError, error:
            log.LogFactory.getLogger().warn(_("File was already closed (ignored): %s") % error)

        return header

    def setHeader(self, timeofBackup):
        """
        Sets the header of the snar File. 
        GNU tar-1.19-2  -> in the first line
        second line is timeofBackupInSec\000timeofBackupInNano
        @param timeofBackup: The time to set in the snar file
        @type timeofBackup: datetime
        """
        if type(timeofBackup) != datetime :
            raise SBException("timeofBackup must be a datetime")
        fd = open(self.snpfile, 'w')

        fd.write("GNU tar-1.19-2\n")
        t = int(time.mktime(timeofBackup.timetuple()))
        fd.write(2 * (str(t) + self.__SEP))
        fd.close()

    def addRecord(self, record):
        """
        Write a record in the snar file. A record is a tuple with 6 entries + a content that is a dict
        @param record: A tuple that contains the record to add. [nfs,mtime_sec,mtime_nano,dev_no,i_no,name,contents] where contents is a dict of {file:'control'}
        """
        if len(record) != 7:
            raise ValueError("Record must contain of 7 elments. Got %s instead." % str(len(record)))

        woContent, contents = record[:-1], record[-1]
        strContent = self.createContent(contents)   # compute contents
        toAdd = "%s%s%s" % (self.__SEP.join(woContent), self.__SEP, strContent)

        fd = _FOP.openfile_for_append(self.snpfile)
        fd.write("%s%s" % (toAdd, self.__entrySEP))
        fd.close()


    def createContent(self, contentList):
        """Creates a content item from a list of given Dumpdirs.
        
        @param contentList: the list of Dumpdirs
        @type contentList: list
        @return: a string containing the computed content
        @rtype: string
        """
        if type(contentList) != list :
            raise SBException("Contentlist must be a list : %r" % repr(contentList))

        result = ""

        for dumpdir in contentList:
            if not isinstance(dumpdir, Dumpdir):
                raise TypeError("Contentlist must contain elements of type "\
                                "'Dumdir'. Got %s instead." % type(dumpdir))
#            result += dumpdir.getControl() + dumpdir.getFilename() + self.__SEP
            result = "%s%s%s%s" % (result, dumpdir.getControl(), dumpdir.getFilename(), self.__SEP)

        return result

    def get_time_of_backup(self):
        """Returns the time of the snapshot (as stored in the snapshot file)
        in seconds since beginning of the epoch.
        
        @rtype: Float
        """
        _header = self.getHeader()
        _header_t = _header.split("\n")
        assert len(_header_t) == 2
        _time_t = _header_t[1].split(self.__SEP)
        assert len(_time_t) > 1
        _time = float("%s.%s" % (_time_t[0], _time_t[1]))
        return _time

    def get_dict_format2(self):
        """Returns the content of the snapshot file as dictionary.
        This provides a much faster way of accessing the content than
        parsing the file on demand. The required RAM is moderate,
        a 10GiB partition filled with files results in 6MiB dictonary.
        
        @warning: only compatible tar version 2 of Tar format
        @note: Uses the same algorithm as method `parseFormat2` and is ~15% faster
                than the high-level variant below.
        """
        _snardict = {}

        fd = _FOP.openfile_for_read(self.snpfile)
#TODO: Handle empty files properly
        header = ""
        n = 0
        while n < 2 :
            c = fd.read(1)
            if len(c) != 1:
                raise SBException(_("The snarfile header is incomplete."))
            if c == '\0':
                n += 1
            header += c
        # skip header which is the first line and 2 entries separated with NULL in this case
#FIXME: remove code duplication here and in `getHeader`

        currentline = ""
        last_c = ''
        c = fd.read(1)
        while c:
            currentline += c
            if c == '\0' and last_c == '\0' :
                # we got a line
                nfs, mtime_sec, mtime_nano, dev_no, i_no, _dirname, \
                    _content = currentline.lstrip("\0").split("\0", 6)
                _snardict[_dirname] = Dumpdir.DIRECTORY
                _content_t = _content.rstrip('\0').split('\0')
                for _entry in _content_t:
                    _entry = _entry.strip("\0")
                    if _entry:
                        _epath = _FOP.joinpath(_dirname, _entry[1:])
                        _snardict[_epath] = _entry[0]

                currentline = ''
                last_c = ''
            else :
                last_c = c
            c = fd.read(1)

        fd.close
        return _snardict

#    def get_dict_format2(self):
#        """        
#        @warning: only compatible tar version 2 of Tar format
#
#        more high-level programmed
#        """
##        _dirdict = {}
#        _snardict = {}
#
#        for _record in self.parseFormat2():
##            print _record
#            _dirname = _record[self.REC_DIRNAME]
#            _content = _record[self.REC_CONTENT]
##            if _dirname in _dirdict:
##                raise ValueError("Directory is already contained.")
##            _dirdict[_dirname] = _content
#            
#            _snardict[_dirname] = Dumpdir.DIRECTORY
#            
#            for _entry in _content:
##                print "Type: %s name: '%s' control: '%s'" %(type(_entry),
##                                                        _entry.getFilename(),
##                                                        _entry.getControl())
#                _epath = fam.joinpath(_dirname, _entry.getFilename())
#                _snardict[_epath] = _entry.getControl()
#            
#
##        print "Dir dictionary:\n%s" % _dirdict              
##        print "Snar dictionary:\n%s" % _snardict  
#        return _snardict


class SnapshotFileWrapper(object):
    """Something like an Interface class.
    
    @todo: Review/Implement
    """
    def __init__(self):
        pass

    def get_snapfile_path(self):
        raise NotImplementedError


class MemSnapshotFile(SnapshotFileWrapper, SBdict):
    """
    This is a representation in memory of a simplified SNAR file. The used structure is an SBDict.
    The "prop" value is the content of the directory. wich is a list of L{Dumpdir}
    
    @note: In the series 0.2 implementation this variant is not used.
    """

    def __init__(self, snapshotFile):
        """
        load the snapshotFile in memory
        @param snapshotFile: a SnapshotFile to convert in MemSnapshotFile
        @type snapshotFile: sbackup.util.tar.SnapshotFile
        """
        if not isinstance(snapshotFile, SnapshotFile) :
            raise Exception("A SnapshotFile is required")

        SnapshotFileWrapper.__init__(self)
#        SBdict.__init__(self, mapping)

        self.__snapshotFile = snapshotFile

        for f in snapshotFile.parseFormat2():
            self[f[-2]] = f[-1]

    def get_snapfile_path(self):
        return self.__snapshotFile.get_filename()

    def hasPath(self, path):
        """
        Checks if a path is include in the SNAR file
        @param path: The path to check
        @return: True if the file is included, False otherwise
        @rtype: boolean
        """
        return self.has_key(path)

    def addRecord(self, record):
        """
        Write a record in the snar file. A record is a tuple with 6 entries + a content that is a dict
        @param record: A tuple that contains the record to add. [nfs,mtime_sec,mtime_nano,dev_no,i_no,name,contents] where contents is a dict of {file:'control'}
        """
        self.__snapshotFile.addRecord(record)
        self[record[-2]] = record[-1]

    def getHeader(self):
        self.__snapshotFile.getHeader()

    def setHeader(self, timeofBackup):
        """
        Sets the header of the snar File. 
        GNU tar-1.19-2  -> in the first line
        second line is timeofBackupInSec\000timeofBackupInNano
        @param timeofBackup: The time to set in the snar file
        @type timeofBackup: datetime
        """
        self.__snapshotFile.setHeader(timeofBackup)

    def getContent(self, dirpath):
        """
        convenance method to get the content of a directory.
        @param dirpath: The directory absolute path to get
        @type dirpath: str
        @return: The content of the dir
        @rtype: list
        """
        return self[dirpath]

    def getFirstItems(self):
        """
        Get the first items in this SnapshotFile (the lower level dirs in the file)
        @return: A list of paths
        @rtype: list 
        """
        result = list()
        for f in self.iterFirstItems():
            result.append(f)
        return result


class ProcSnapshotFile(SnapshotFileWrapper):
    """This is a Snapshotfile that will basically every time parse the
    snarfile for information.
    """

    def __init__(self, snapshotFile):
        """
        load the snapshotFile to get a reference on it
        @param snapshotFile: a SnapshotFile to convert in MemSnapshotFile
        @type snapshotFile: sbackup.util.tar.SnapshotFile
        """
        if not isinstance(snapshotFile, SnapshotFile) :
            raise TypeError(_("A SnapshotFile is required"))

        SnapshotFileWrapper.__init__(self)

        self.__snapshotFile = snapshotFile

    def __str__(self):
        _str = "Snar file: '%s'" % self.__snapshotFile
        return _str

    def get_snapfile_path(self):
        return self.__snapshotFile.get_filename()

    def get_snapfile_obj(self):
        """Returns the wrapped instance of the snapshot file.
        
        @return: snapshot file object
        @rtype: SnapshotFile
        """
        return self.__snapshotFile

    def hasPath(self, path):
        """
        Checks if a path is included in the SNAR file
        @param path: The path to check
        @return: True if the file is included, False otherwise
        @rtype: boolean
        """
        for f in self.__snapshotFile.parseFormat2():
            if f[SnapshotFile.REC_DIRNAME].rstrip(_FOP.pathsep) == path.rstrip(_FOP.pathsep):
                return True
        return False

    def hasFile(self, _file):
        """
        Checks if a file is included in the SNAR file. a file is in a directory thus in the content.
        @param _file: The file to check. a complete path must be given
        @return: True if the file is included, False otherwise
        @rtype: boolean
        """
        dir, inFile = _file.rsplit(_FOP.pathsep, 1)
        if not self.hasPath(dir):
            return False
        for f in self.getContent(dir):
            if f.getFilename() == inFile and f.getControl() != Dumpdir.UNCHANGED:
                return True
        return False

    def iterfiles(self):
        """
        iterator that returns every file in the snapshot
        """
        for l in self.__snapshotFile.parseFormat2():
            dir = l[-2]
            if l[-1] :
                for dumpdir in l[-1] :
                    yield dir + _FOP.pathsep + dumpdir.getFilename()

    def iterRecords(self):
        """Iterator over snar file records (wrapper on the
        'parseFormat2' method of SnaspshotFile
        """
        for record in self.__snapshotFile.parseFormat2():
            yield record


    def addRecord(self, record):
        """
        Write a record in the snar file. A record is a tuple with 6 entries + a content that is a dict
        @param record: A tuple that contains the record to add. [nfs,mtime_sec,mtime_nano,dev_no,i_no,name,contents] where contents is a dict of {file:'control'}
        """
        self.__snapshotFile.addRecord(record)

    def getHeader(self):
        return self.__snapshotFile.getHeader()

    def setHeader(self, timeofBackup):
        """
        Sets the header of the snar File. 
        GNU tar-1.19-2  -> in the first line
        second line is timeofBackupInSec\000timeofBackupInNano
        @param timeofBackup: The time to set in the snar file
        @type timeofBackup: datetime
        """
        self.__snapshotFile.setHeader(timeofBackup)

    def getContent(self, dirpath):
        """convenance method to get the content of a directory.
        
        @param dirpath: The directory absolute path to get
        @type dirpath: str
        
        @return: The content of the dir
        @rtype: list
        
        @raise SBException: if the path isn't found in the snapshot file
        """
        for f in self.__snapshotFile.parseFormat2():
            if f[SnapshotFile.REC_DIRNAME].rstrip(_FOP.pathsep) == dirpath.rstrip(_FOP.pathsep):
                return f[SnapshotFile.REC_CONTENT]
        raise SBException(_("Directory does not exist: %s.") % dirpath)

    def getFirstItems(self):
        """
        Get the first items in this SnapshotFile (the lower level dirs in the file)
        @return: A list of paths
        @rtype: list 
        """

        def cleanDupl():
            if result:
                result.sort()
                last = result[-1]
                for i in range(len(result) - 2, -1, -1):
                    if last == result[i]:
                        del result[i]
                    else:
                        last = result[i]

        result = list()

        for f in self.__snapshotFile.parseFormat2():
            found = False

            for i in range(0, len(result)) :
                if result[i] == f[-2] or f[-2].startswith(result[i]) :
                    if found :
                        # then we are trying to overide
                        continue
                    else :
                        found = True
                        break
                elif result[i].startswith(f[-2]):
                    # replace with f[-2]
                    result[i] = f[-2]
                    found = True
                    # don't break cause it's possible that some others entries need to be overiden

            if not found :
                result.append(f[-2])
            else :
                cleanDupl()

        return result



#   Simple Backup - Basic data structures
#
#   Copyright (c)2009-2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import os

from sbackup.util.exceptions import SBException
from sbackup.util.exceptions import CorruptedSBdictException


class Singleton(type):
    def __init__(cls, name, bases, dic): #IGNORE:C0203
        type.__init__(cls, name, bases, dic)
        cls.instance = None

    def __call__(self, *args, **kw): #IGNORE:C0203
        if self.instance is None:
            self.instance = type.__call__(self, *args, **kw)
        return self.instance


class SBdict(dict) :
    """
    This is a structure used by sbackup to store flist and fprops . 
    The structure is in fact a tree meant to :
    - ease the retrieval of the content of a folder
    - to optimise the use of memory to store the flist file by avoiding to store identical subpath.
    
    The flist file has this contents :
    
    /home
    /home/wattazoum
    /home/wattazoum/Desktop
    /home/wattazoum/Desktop/sbackup-test
    /home/wattazoum/Desktop/sbackup-test/d17
    /home/wattazoum/Desktop/sbackup-test/d17/d3
    /home/wattazoum/Desktop/sbackup-test/d17/d3/f4.txt
    /home/wattazoum/Desktop/sbackup-test/d17/d2
    /home/user/Desktop/sbackup-test/d17/d3/f4.txt
    
    The goal of SBdict is to store the data this way
    
    /home    ->wattazoum    ->Desktop    ->sbackup-test    ->d17    ->d3    ->f4.txt
                                                            ->d2
            ->user        ->Desktop    ->sbackup-test    ->d17    ->d3    ->f4.txt
    
    We use for that this node :
    { 'file' : [ 'props', sonNode ] }
    
    Example : 
    
    { 'home' : [ 'props', {
         'wattazoum' : [ 'props', { 
             'Desktop' : [ 'props', { 
                 'sbackup-test' : [ 'props', { 
                     'd17' : [ 'props', { 
                         'd3' : [ 'props', { 
                             'f4.txt' : [ 'props', None ] } ] ,
                         'd2' : [ 'props', None ] } ] } ] } ] } ] , 
         'user' : [ 'props', { 
             'Desktop' : [ 'props', {
                 'sbackup-test' : [ 'props', { 
                     'd17' : [ 'props', { 
                         'd3' : [ 'props', { 
                             'f4.txt' : [ 'props', None ] } ] } ] } ] } ] } ] 
         } ] } 
    
    We can save the memory for all subpathes but then, we might have a lost of speed when
    doing some search (comparing to putting the whole file in a Dictionary).
    
    @note: And yes, this penalty in speed is notable!
           We should replace it or at least provide a faster alternative.
    """
    def __init__(self, mapping = None):
        if not mapping :
            dict.__init__(self)
        else :
            if type(mapping) == list and type(mapping[0]) == tuple and len(mapping[0]) == 2 :
                dict.__init__(self)
                for key, value in mapping :
                    self.__setitem__(key, value)
            else :
                raise Exception ("Not implemented yet")

#    def __str__(self):
#        _out = []
#        for _item in self:
#            _out.append("%s" % _item)
#        return "\n".join(_out)

    def getSon(self, path) :
        """
        get the son SBdict of a path
        @param path: the path to get the son of
        @return: a SBdict or None if there were no son
        """
        if self.has_key(path) :
            return self[path][1]
        else :
            raise SBException("'%s' not in SBdict" % path)

    def setSon(self, path, son):
        """
        Set the son of a path 
        (/!\ the props of the path will be set to None if the key didn't exist )
        @param path: the path to set the son on
        @param son: the son as a SBdict  
        """
        if son != None and type(son) != SBdict :
            raise CorruptedSBdictException("You can't set '%s' as a son " % str(son))
        if not self[path] or self[path][0] == None :
            self.__setitem__(path, [None, son])
        else :
            self.__setitem__(path, [self[path][0], son])

    def has_key(self, key) :
        """
        Return True if the path have been found and false if not
        @param key: a path to search (/home/user/test/dir )
        """

        splited = key.split(os.sep, 1)

        if not dict.has_key(self, splited[0]) :
            return False
        else : # base dir found, we look for the son
            # if the son is empty, we found our element
            if len(splited) == 1 or not splited[1] :
                return True
            if self[splited[0]][1] != None :
                if type(self[splited[0]][1]) == SBdict :
                    return self[splited[0]][1].has_key(splited[1])
                else :
                    raise CorruptedSBdictException("The value stored in the SBdict is Invalid : " + str(type(self[splited[0]][1])))
            else : return False

    def __setitem__(self, key, value) :
        """
        Add an item
        @param key: a string 
        @param value: Value must be None , a String or a 2 length list with None or an SBdict on the second member
        @raise CorruptedSBdictException: 
        """
        valIsSubtree = False

        if value != None :
            if type(value) == list and len(value) == 2 and (value[1] is None or (value[1] != None and type(value[1]) == SBdict))  :
                valIsSubtree = True

        splited = key.split(os.sep, 1)

        if len(splited) == 1 or not splited[1] :
            # we are at the end of a path 
            # we fallback to the normal behaviour
            if dict.has_key(self, splited[0]) :
                # The key exists
                item = dict.__getitem__(self, splited[0])
                if not valIsSubtree :
                    prop = value
                    dict.__setitem__(self, splited[0], [prop, item[1]])
                else :
                    prop = value[0]
                    if value[1] == None :
                        dict.__setitem__(self, splited[0], [prop, item[1]])
                    else :
                        dict.__setitem__(self, splited[0], value)

            else :
                #the key doesn't exist
                if not valIsSubtree :
                    dict.__setitem__(self, splited[0], [value, None])
                else :
                    dict.__setitem__(self, splited[0], value)
        else : # path is composed , 
            # we check if the base dir exists and get the props infos
            if dict.has_key(self, splited[0]) :
                item = dict.__getitem__(self, splited[0])
                prop = item[0]
                if item[1] != None :
                    if type(item[1]) != SBdict :
                        raise CorruptedSBdictException("The value stored in the SBdict is Invalid : " + str(item[1]))
                    son = item[1]
                else :
                    son = SBdict()
            else:
                # get the properties of the base dir
                #s = os.lstat(splited[0])
                #prop = str(s.st_mode)+str(s.st_uid)+str(s.st_gid)+str(s.st_size)+str(s.st_mtime)
                prop = None
                son = SBdict()
            son.__setitem__(splited[1], value)
            dict.__setitem__(self, splited[0], [prop, son])

    def __delitem__(self, key):
        """
        """
        if not self.has_key(key) : return False

        if key == os.sep :
            dict.__delitem__(self, "")
            return True
        else :
            spl = key.rstrip(os.sep).split(os.sep)
            nkey = os.sep.join(spl[:len(spl) - 1])
            last = spl[len(spl) - 1]
            dict.__delitem__(self.getSon(nkey), last)
            return True


    def __getitem__(self, key) :
        """
        Return the item Value
        @param key: a path to search (/home/user/test/dir )
        @return: False if the item is not found
        """
        splited = key.split(os.sep, 1)

        if not dict.has_key(self, splited[0]) :
            return False
        else : # base dir found, we look for the son
            # if the son is empty, we found our element
            if len(splited) == 1 or not splited[1] :
                return dict.__getitem__(self, splited[0])
            if self[splited[0]][1] != None :
                if type(self[splited[0]][1]) == SBdict :
                    return self[splited[0]][1].__getitem__(splited[1])
                else :
                    raise CorruptedSBdictException("The value stored in the SBdict is Invalid : " + str(type(self[splited[0]][1])))
            else : return False

    def iterkeys(self, _path = None) :
        """Returns an iterator that goes recursively through the full paths.
        
        Should return fullpath (means what?)
        """
        if _path is None: # initialization
            _path = []
        for dirname, (props, son) in dict.iteritems(self):
            _path.append(dirname)
            yield os.sep.join(_path)
            if son is None:
                _path.pop()
            else:
                for path in son.iterkeys(_path):
                    yield path
        if len(_path) > 0 :
            _path.pop()

    def iteritems(self, _path = None) :
        """Iterator that goes recursively through the whole dictionary and returns
        paths and their properties. Every sub-path is considered.
        
        @return: (fullpath, props)
        """
        if _path is None: # initialization
            _path = []
        for dirname, (props, son) in dict.iteritems(self):
            _path.append(dirname)
            yield (os.sep.join(_path), props)
            if son is None:
                _path.pop()
            else:
                for path, prop in son.iteritems(_path):
                    yield (path, prop)
        if len(_path) > 0 :
            _path.pop()

    def itervalues(self, _path = None) :
        """
        an Iterator that gets recursively the full path
        Should return props
        """
        if _path is None: # initialization
            _path = []
        for dirname, (props, son) in dict.iteritems(self):
            _path.append(dirname)
            yield props
            if son is None:
                _path.pop()
            else:
                for prop in son.itervalues(_path):
                    yield prop
        if len(_path) > 0 :
            _path.pop()

    def iterFirstItems(self, _path = None):
        """
        an Iterator that gets recursively the path off first items
        Should return props
        """
        if _path is None: # initialization
            _path = []
        for dirname, (props, son) in dict.iteritems(self):
            _path.append(dirname)
            if props != None :
                yield os.sep.join(_path)
                _path.pop()
            else:
                if son is not None :
                    for path in son.iterFirstItems(_path):
                        yield path
                else :
                    raise SBException("getting to an ending file without properties")
        if len(_path) > 0 :
            _path.pop()

    def getEffectiveFileList(self, path = None):
        """Iterator that returns the effective files list. Effective means that all 'end-nodes' or
        leafs of the tree are returned. Unlike method `iterkeys` are *not* all sub-paths returned.        
        Paths are included in the effective list of files if their `props` are set (i.e. set to 0 or 1,
        not set to None) since this is not true for sub-paths. Moreover, paths can be 'disabled' (i.e.
        excluded from the effective list of files) by setting their `props` to None.
        """
        for _file, _prop in self.iteritems(path):
            if _prop is not None:
                yield _file

    def get_eff_filelist_not_nested(self, path = None):
        """
        an Iterator that return the effective files list to give to TAR.
        This means, some files won't appear because they are part of an included sub directory
        """
        for _file in self.getEffectiveFileList(path):
            if not self.hasParentDirIncluded(_file) :
                yield _file

    def hasFile(self, _file):
        """Checks if the SBdict has a file. Unlike has_key, this will not match subdirectories name
        """
        if not self.has_key(_file) :
            return False
        else :
            if self.__getitem__(_file)[0] is None :
                return False
            else :
                return True

    def contains_path(self, path):
        """Checks whether the given `path` is stored in this SBDict. Unlike `hasFile` this
        will also match sub-directories. 
        """
        if self.has_key(path) :
            return True
        return False

    def hasParentDirIncluded(self, path):
        """
        Checks for a path if we have an effective parent dir
        @param path: The path to check 
        @type path: like /d/d1/d2
        """
        if not path:
            return False

        splited = path.rsplit(os.sep, 1)
        _basename = splited[0]
        if self.__getitem__(_basename)[0] is not None:
            return True
        else:
            return self.hasParentDirIncluded(_basename)



#   Simple Backup - GTK+ common helper functionality
#
#   Copyright (c)2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

"""Module containing commonly used helper classes and functions related to UI
"""

from gettext import gettext as _

import gtk
import glib
import gobject
import pango
import types
import traceback


from sbackup.pkginfo import Infos
from sbackup.util import constants
from sbackup.util import system
from sbackup.util import log
from sbackup import util


MODEL_COLUMN_INDEX_VALUE = 0
MODEL_COLUMN_INDEX_KEY = 1


def set_model(widget, values):
    """
    sorted by dictionary keys
    """
    if not isinstance(values, types.DictionaryType):
        raise TypeError

    # query types
    _entry = values.popitem()
    values[_entry[0]] = _entry[1]    # re-add removed pair
    _keytype = type(_entry[0])
    _valtype = type(_entry[1])

    # ensure all dictionary entries are of same type
    for _entry in values:
        if not isinstance(_entry, _keytype):
            raise TypeError
        if not isinstance(values[_entry], _valtype):
            raise TypeError

    _model = gtk.ListStore(_valtype, _keytype)
    _keys = values.keys()
    _keys.sort()

    for _key in _keys:
        _model.append([values[_key], _key])

    widget.set_model(_model)
    _cell = gtk.CellRendererText()
    widget.pack_start(_cell, True)
    widget.add_attribute(_cell, 'text', MODEL_COLUMN_INDEX_VALUE)
    return _model


def open_uri(uri, timestamp = 0):
    """Convenience function for opening a given URI with the default
    application.
    """
    # The function 'show_uri()' is available in PyGTK 2.14 and above.
    if gtk.pygtk_version >= (2, 14, 0):
        gtk.show_uri(gtk.gdk.screen_get_default(), uri, timestamp)
    else:
        try:
            import gnome
            gnome.url_show(uri)
        except ImportError:
            pass


def label_set_autowrap(widget):
    """Make labels automatically re-wrap if their containers are resized.
    Accepts label or container widgets.    For this to work the label in the
    glade file must be set to wrap on words.
    
    from http://stackoverflow.com/questions/1893748/pygtk-dynamic-label-wrapping
    """
    if isinstance(widget, gtk.Container):
        children = widget.get_children()
        for i in xrange(len(children)):
            label_set_autowrap(children[i])
    elif isinstance(widget, gtk.Label) and widget.get_line_wrap():
        widget.connect_after("size-allocate", _label_size_allocate)

def _label_size_allocate(widget, allocation):
    """Callback which re-allocates the size of a label.

    from http://stackoverflow.com/questions/1893748/pygtk-dynamic-label-wrapping
    """
    layout = widget.get_layout()
    lw_old, lh_old = layout.get_size()
    # fixed width labels
    if lw_old / pango.SCALE == allocation.width:
        return
    # set wrap width to the pango.Layout of the labels
    layout.set_width(allocation.width * pango.SCALE)

    # law is unused.
    law, lah = layout.get_size() #IGNORE:W0612
    if lh_old != lah:
        widget.set_size_request(-1, lah / pango.SCALE)

def show_infodialog(message_str, parent, headline_str = "", secmsg_str = ""):
    """Creates und displays a modal dialog box. Main purpose is
    displaying of error messages.
    
    @param message_format: error message to show
    @type message_format: String
    
    """
    __show_msgdialog(message_str = message_str, msgtype = gtk.MESSAGE_INFO,
                    parent = parent, boxtitle = "",
                    headline_str = headline_str, secmsg_str = secmsg_str)

def show_warndialog(message_str, parent, headline_str = "", secmsg_str = ""):
    """Creates und displays a modal dialog box. Main purpose is
    displaying of error messages.
    
    @param message_format: error message to show
    @type message_format: String
    
    """
    __show_msgdialog(message_str = message_str, msgtype = gtk.MESSAGE_WARNING,
                    parent = parent, boxtitle = "",
                    headline_str = headline_str, secmsg_str = secmsg_str)

def show_errdialog(message_str, parent, headline_str = "", secmsg_str = ""):
    """Creates und displays a modal dialog box. Main purpose is
    displaying of error messages.
    
    @param message_format: error message to show
    @type message_format: String
    """
    __show_msgdialog(message_str = message_str, msgtype = gtk.MESSAGE_ERROR,
                    parent = parent, boxtitle = "",
                    headline_str = headline_str, secmsg_str = secmsg_str)

def show_errdialog_threaded(message_str, parent, headline_str = "", secmsg_str = ""):
    gtk.gdk.threads_enter()
    __show_msgdialog(message_str = message_str, msgtype = gtk.MESSAGE_ERROR,
                    parent = parent, boxtitle = "",
                    headline_str = headline_str, secmsg_str = secmsg_str)
    gtk.gdk.threads_leave()

def __show_msgdialog(message_str, msgtype, parent, boxtitle = "",
                    headline_str = "", secmsg_str = ""):
    """Creates und displays a modal dialog box. Main purpose is
    displaying of error messages.
    
    Do not use markup for the strings.
    
    @param message_format: error message to show
    @type message_format: String
    
    @todo: Add proper escaping before markup is applied to the headlline.
    """
    # in compliance with Gnome HIG a 'Close' button instead of 'OK' is used

    dialog = msgdialog(message_str = message_str, msgtype = msgtype, parent = parent, boxtitle = boxtitle,
                         buttons = gtk.BUTTONS_CLOSE, headline_str = headline_str, secmsg_str = secmsg_str)

    # the message box is showed
    dialog.run()
    dialog.destroy()

def msgdialog(message_str, msgtype, parent, boxtitle = "", buttons = gtk.BUTTONS_CLOSE,
                    headline_str = "", secmsg_str = ""):
    """Creates und displays a modal dialog box. Main purpose is
    displaying of error messages.
    
    Do not use markup for the strings.
    
    @param message_format: error message to show
    @type message_format: String
    
    @todo: Add proper escaping before markup is applied to the headlline.
    """
    # in compliance with Gnome HIG a 'Close' button instead of 'OK' is used

    dialog = gtk.MessageDialog(parent = parent, type = msgtype,
                flags = gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                buttons = buttons)

    icon_file = util.get_resource_file(constants.DEFAULT_ICON_FILENAME)
    dialog.set_icon_from_file(icon_file)

    if boxtitle.strip() != "":
        dialog.set_title(boxtitle)

    _hdl = headline_str.strip(" \n\t")
    if _hdl != "":
        _hdl = "<b>%s</b>\n\n" % glib.markup_escape_text(_hdl)
    _msg = "%s%s" % (_hdl, glib.markup_escape_text(message_str))
    dialog.set_markup(_msg)

    # an optional secondary message is added
    _sec = secmsg_str.strip(" \n\t")
    if _sec != "":
        _sec = "<small>%s</small>" % (glib.markup_escape_text(_sec))
        dialog.format_secondary_markup(_sec)
    return dialog


def infodialog_standalone(message_str, boxtitle, headline_str = "", secmsg_str = "", sticky = True):
    return msgdialog_standalone(message_str = message_str, msgtype = gtk.MESSAGE_INFO,
                                  boxtitle = boxtitle, headline_str = headline_str, secmsg_str = secmsg_str,
                                  sticky = sticky)


def warndialog_standalone(message_str, boxtitle, headline_str = "", secmsg_str = "", sticky = True):
    return msgdialog_standalone(message_str = message_str, msgtype = gtk.MESSAGE_WARNING,
                                  boxtitle = boxtitle, headline_str = headline_str, secmsg_str = secmsg_str,
                                  sticky = sticky)


def errdialog_standalone(message_str, boxtitle, headline_str = "", secmsg_str = "", sticky = True):
    return msgdialog_standalone(message_str = message_str, msgtype = gtk.MESSAGE_ERROR, boxtitle = boxtitle,
                                  headline_str = headline_str, secmsg_str = secmsg_str, sticky = sticky)


def msgdialog_standalone(message_str, msgtype, boxtitle, buttons = gtk.BUTTONS_CLOSE,
                    headline_str = "", secmsg_str = "", sticky = True):
    """Creates and displays a standalone dialog box. Main purpose is
    displaying of error messages within the indicator application.
    
    Do not use markup for the strings.
    
    @param message_format: error message to show
    @type message_format: String
    
    @todo: Add proper escaping before markup is applied to the headlline.
    """
    # in compliance with Gnome HIG a 'Close' button instead of 'OK' is used

    assert boxtitle.strip() != ""

    dialog = msgdialog(message_str = message_str, msgtype = msgtype, parent = None, boxtitle = boxtitle,
                       buttons = buttons, headline_str = headline_str, secmsg_str = secmsg_str)

    dialog.set_property("skip-pager-hint", False)
    dialog.set_property("skip-taskbar-hint", False)

    if sticky is True:
        dialog.stick()
        dialog.set_urgency_hint(True)
#        dialog.set_keep_above(True) # be not to urgent

    return dialog


def show_about_dialog(set_transient_for = None):
    _icon_file = util.get_resource_file(constants.DEFAULT_ICON_FILENAME)
    about = gtk.AboutDialog()
    about.set_name(Infos.NAME)
    about.set_version(Infos.VERSION)
    about.set_comments(Infos.DESCRIPTION)
    if set_transient_for is not None:
        about.set_transient_for(set_transient_for)
    about.set_copyright(Infos.COPYRIGHT)
    about.set_translator_credits(Infos.TRANSLATORS)
    about.set_authors(Infos.AUTHORS)
    if not system.is_superuser():
        about.set_website(Infos.WEBSITE)
    about.set_logo(gtk.gdk.pixbuf_new_from_file(_icon_file))
    about.set_icon_from_file(_icon_file)
    about.run()
    about.destroy()


def get_statusbar_msg_mode(statusmessage):
    if system.is_superuser():
        stattxt = _("%s   (Administrator mode)") % statusmessage
    else:
        stattxt = statusmessage
    return stattxt


def set_watch_cursor(widget):
    watch = gtk.gdk.Cursor(gtk.gdk.WATCH)
    try:
        widget.window.set_cursor(watch)
    except AttributeError:
        pass
#        print "Ignored error when setting cursor"


def unset_cursor(widget):
    try:
        widget.window.set_cursor(None)
    except AttributeError:
        pass
#        print "Ignored error when un-setting cursor"


def _common_except_hook(etype, evalue, etb):
    _logger = log.LogFactory.getLogger()
    _lines = traceback.format_exception(etype, evalue, etb)
    _lines = "".join(_lines)
    _lines.strip()
    _logger.error("Uncaught exception: %s" % evalue)
    _logger.error(_lines)
    _message_str = _("An uncaught error occurred. Close this message window and restart the application.\n\nPlease report this error on https://bugs.launchpad.net/sbackup.\n\n%s") % (evalue)
    _headline_str = _("Sorry, this should not have happened")
    _secmsg_str = "%s" % _lines
    return (_headline_str, _message_str, _secmsg_str)


def except_hook(etype, evalue, etb):
    _headline_str, _message_str, _secmsg_str = _common_except_hook(etype, evalue, etb)
    show_errdialog(message_str = _message_str,
                     parent = None, headline_str = _headline_str,
                     secmsg_str = _secmsg_str)


def except_hook_threaded(etype, evalue, etb):
    _headline_str, _message_str, _secmsg_str = _common_except_hook(etype, evalue, etb)
    gobject.idle_add(show_errdialog_threaded, _message_str,
                     None, _headline_str,
                     _secmsg_str)



#   Simple Backup - Indicator application (status icon)
#                   core implementation
#
#   Copyright (c)2009-2010,2012,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import os
import time

from gettext import gettext as _

import gobject
import glib
import dbus.mainloop.glib
import gtk


from sbackup.pkginfo import Infos
from sbackup import util
from sbackup.util import dbus_support
from sbackup.util.log import LogFactory
from sbackup.util import constants
from sbackup.util import system
from sbackup.util import lock
from sbackup.util import exceptions
from sbackup.ui import misc


class INotifyMixin(object):
    """Mix-in class that provides the displaying of notifications using the
    pynotify module.
    
    """
    def __init__(self, logger, iconfile, trayicon = None):
        pass

    def _notify_info(self, profilename, message):
        """Shows up a pop-up window to inform the user. The notification
        supports mark-up.        

         :param message: The message (body) that should be displayed.
         :type message:  String
         
        """
        raise NotImplementedError

    def _notify_warning(self, profilename, message):
        """Shows up a pop-up window to inform the user. The notification
        supports mark-up.        

         :param message: The message (body) that should be displayed.
         :type message:  String
         
        """
        raise NotImplementedError

    def _notify_error(self, profilename, message):
        """Shows up a pop-up window to inform the user that an error occured.
        Such error notifications are emphasized and must be closed manual. The
        notifications support mark-up.

         :param message: The message (body) that should be displayed.
         :type message:  String
         
        """
        raise NotImplementedError


class PyNotifyMixin(INotifyMixin):
    """Mix-in class that provides the displaying of notifications using the
    pynotify module.    
    """
    def __init__(self, logger, iconfile, trayicon = None):
        """Default constructor.
        
        :param logger: Instance of logger to be used.
        
        :todo: The notification domain should be retrieved from a central place!
        
        """
        INotifyMixin.__init__(self, logger, iconfile, trayicon)
        self.__logger = logger
        self.__trayicon = trayicon
        self.__iconfile = iconfile

        # internal flag whether the notification module is usable
        self.__pynotif_avail = False

        # the pynotify module is stored in this variable
        self.__pynotif_mod = None

        # the current notification
        self.__notif = None

        # trying to initialize the notification module
        try:
            import pynotify
            self.__pynotif_mod = pynotify
            if self.__pynotif_mod.init(constants.NOTIFICATION_DOMAIN):
                self.__pynotif_avail = True
            else:
                self.__pynotif_avail = False    # yes, this is insane!
                self.__logger.warning(_("there was a problem initializing the "\
                                        "pynotify module"))
        except ImportError, exc:
            self.__pynotif_avail = False
            self.__logger.warning(str(exc))


    def _notify_info(self, profilename, message):
        """Shows up a pop-up window to inform the user. The notification
        supports mark-up.        

         :param message: The message (body) that should be displayed.
         :type message:  String
         
        """
        if self.__pynotif_avail:
            if self.__notif is None:
                self.__notif = self.__get_notification(profilename, message)
            else:
                self.__update_notification(profilename, message)

            if isinstance(self.__notif, self.__pynotif_mod.Notification):
                try:
                    if self.__trayicon is not None:
                        try:
                            self.__notif.attach_to_status_icon(self.__trayicon)
                        except AttributeError: pass
                    self.__notif.set_urgency(self.__pynotif_mod.URGENCY_LOW)
                    self.__notif.show()
                except gobject.GError, exc:
                    # Connection to notification-daemon failed 
                    self.logger.warning(_("Connection to notification-daemon failed: %s") % str(exc))

    def _notify_warning(self, profilename, message):
        """Shows up a pop-up window to inform the user. The notification
        supports mark-up.        

         :param message: The message (body) that should be displayed.
         :type message:  String
         
        """
        self.__notify_new(profilename, message, mode = "warning")

    def _notify_error(self, profilename, message):
        """Shows up a pop-up window to inform the user that an error occured.
        Such error notifications are emphasized and must be closed manual. The
        notifications support mark-up.

         :param message: The message (body) that should be displayed.
         :type message:  String
         
        """
        self.__notify_new(profilename, message, mode = "critical")

    def __notify_new(self, profilename, message, mode):
        """Shows up a *new* pop-up window to inform the user that an error occured.
        Such error notifications are emphasized and must be closed manual. The
        notifications support mark-up.

        :param message: The message (body) that should be displayed.
        :type message:  String
         
        :note: Comply with Ubuntu Notification Guidelines (no actions, no permanent notifications) 
        """
        if self.__pynotif_avail:
            notif = self.__get_notification(profilename, message)
            if isinstance(notif, self.__pynotif_mod.Notification):
                try:
                    if self.__trayicon is not None:
                        try:
                            notif.attach_to_status_icon(self.__trayicon)
                        except AttributeError: pass

                    if mode == "critical":
                        notif.set_urgency(self.__pynotif_mod.URGENCY_CRITICAL)
                    else:
                        notif.set_urgency(self.__pynotif_mod.URGENCY_NORMAL)
                    notif.show()
                except gobject.GError, exc:
                    # Connection to notification-daemon failed 
                    self.logger.warning(_("Connection to notification-daemon failed: %s") % str(exc))

    def __get_notification(self, profilename, message):
        """Returns a notification object but does not display it. The
        notification supports mark-up. If notifications aren't supported
        the method returns None.
         
        :param message: The message (body) that should be displayed.
        :type message:  String
         
        :return: The created notification object or None
        :rtype: Notification or None
        
        :todo: Replace single '<' characters by '&lt;' in a more reliable way!\
               See function `gobject.markup_escape_text` for this.
        :todo: The header and the icon should be given as parameter to make
               this mix-in class more generic!
               
        """
        notif = None
        if self.__pynotif_avail:
            message = message.replace("<", "&lt;")
            title = self.__get_notification_title(profilename)

            try:
                notif = self.__pynotif_mod.Notification(title, message, self.__iconfile)
            except gobject.GError, exc:
                # Connection to notification-daemon failed 
                self.logger.warning(_("Connection to notification-daemon failed: %s") % str(exc))
                notif = None
        return notif

    def __update_notification(self, profilename, message):
        """         
        :param message: The message (body) that should be displayed.
        :type message:  String
         
        :todo: Replace single '<' characters by '&lt;' in a more reliable way!
        :todo: The header and the icon should be given as parameter to make
               this mix-in class more generic!
               
        """
        if self.__pynotif_avail:
            message = message.replace("<", "&lt;")
            title = self.__get_notification_title(profilename)

            try:
                self.__notif.update(title, message, self.__iconfile)
            except gobject.GError, exc:
                # Connection to notification-daemon failed 
                self.logger.warning(_("Connection to notification-daemon failed: %s") % str(exc))
                self.__notif = None

    def __get_notification_title(self, profilename):
        title = Infos.NAME
        if profilename != "":
            title += "\n[%s]" % profilename
        return title


class SBackupdIndicatorBase(INotifyMixin):

    def __init__(self, indicator_hdl):
        if not isinstance(indicator_hdl, SBackupdIndicatorHandler):
            raise TypeError("Parameter of type `SBackupdIndicatorHandler` expected.")
        self.logger = LogFactory.getLogger()
        INotifyMixin.__init__(self, logger = self.logger, iconfile = None, trayicon = None)

        self._indicator_hdl = indicator_hdl
        self._mainloop = gobject.MainLoop()

        self._indicator = None

        self._exit = False
#TODO: Collect required named dialogs in dictionary.
        self._targetnotfound_dialog = None
        self._cancel_dialog = None
        self._current_dialogs = []
        self._menu = gtk.Menu()
        self._menuitems = {}

        self._connect_dbus_signal_handlers()
        self._init_dbus_check_timer()
        self._init_autoexitcheck_timer()

    def _notify_info(self, profilename, message):
        raise NotImplementedError

    def _notify_warning(self, profilename, message):
        raise NotImplementedError

    def _notify_error(self, profilename, message):
        raise NotImplementedError

    def _init_autoexitcheck_timer(self):
        if self._indicator_hdl.get_keep_alive() is False:
            gobject.timeout_add_seconds(constants.AUTOEXIT_CHECK_INTERVAL_SECONDS,
                                self._autoexitcheck)

    def _init_autoexit_timer(self):
        gobject.timeout_add_seconds(constants.AUTOEXIT_TIMEOUT_SECONDS,
                            self._autoexit)

    def _init_dbus_check_timer(self):
        gobject.timeout_add_seconds(constants.DBUS_CHECK_INTERVAL_SECONDS, self._dbus_check)

    def _init_dbus_reconnect_timer(self):
        gobject.timeout_add_seconds(constants.DBUS_RECONNECT_INTERVAL_SECONDS, self._dbus_reconnect)


    def _build_menu(self):
        _status_msg = self._indicator_hdl.get_unknown_menu_label()
        _menuitems = ({ "name"      : "title",
                        "title"     : _("Simple Backup"),
                        "type"      : "MenuItem",
                        "sensitive" : True,
                        "handler"   : { "activate" : self.on_about_clicked }},

                      { "name"      : "show_windows",
                        "title"     : _("Show message windows"),
                        "type"      : "MenuItem",
                        "sensitive" : True,
                        "handler"   : { "activate" : self.on_showdialogs_clicked }},

                      { "name"      : "separator_1",
                        "title"     : None,
                        "type"      : "SeparatorMenuItem",
                        "sensitive" : None,
                        "handler"   : None },

                      { "name"      : "profile",
                        "title"     : _status_msg["profile"],
                        "type"      : "MenuItem",
                        "sensitive" : False,
                        "handler"   : None },

                      { "name"      : "size_of_backup",
                        "title"     : _status_msg["size_of_backup"],
                        "type"      : "MenuItem",
                        "sensitive" : False,
                        "handler"   : None },

                      { "name"      : "progress",
                        "title"     : _status_msg["progress"],
                        "type"      : "MenuItem",
                        "sensitive" : False,
                        "handler"   : None },

                      { "name"      : "remaining_time",
                        "title"     : _status_msg["remaining_time"],
                        "type"      : "MenuItem",
                        "sensitive" : False,
                        "handler"   : None },

                      { "name"      : "separator_2",
                        "title"     : None,
                        "type"      : "SeparatorMenuItem",
                        "sensitive" : None,
                        "handler"   : None },

                      { "name"      : "cancel",
                        "title"     : _("Cancel Backup"),
                        "type"      : "MenuItem",
                        "sensitive" : False,
                        "handler"   : { "activate" : self.on_cancel_clicked }}
                     )

        for _item in _menuitems:
            if _item["type"] == "MenuItem":
                _menuitem = gtk.MenuItem(_item["title"])

                if _item["handler"] is not None:
                    for _signal in _item["handler"]:
                        _menuitem.connect(_signal, _item["handler"][_signal])

                _menuitem.set_sensitive(_item["sensitive"])
                self._menu.append(_menuitem)
                self._menuitems[_item["name"]] = _menuitem

            elif _item["type"] == "SeparatorMenuItem":
                _sep = gtk.SeparatorMenuItem()
                self._menu.append(_sep)
                self._menuitems[_item["name"]] = _sep

            else:
                raise ValueError("Unknown item type '%s'." % _item["type"])

    def _set_menuitems_status(self, func, *args):
        _status_msg = func(*args)
        for _item in _status_msg:
            _label = _status_msg[_item]
            self._menuitems[_item].set_label(_label)

    def _set_menuitems_status_finished(self):
        self._set_menuitems_status(self._indicator_hdl.get_finished_menu_label)

    def _set_menuitems_status_prepare(self):
        self._set_menuitems_status(self._indicator_hdl.get_prepare_menu_label)

    def _set_menuitems_status_unknown(self):
        self._set_menuitems_status(self._indicator_hdl.get_unknown_menu_label)

    def _set_menuitems_status_canceled(self):
        self._set_menuitems_status(self._indicator_hdl.get_canceled_menu_label)

    def _set_menuitems_status_progress(self, checkpoint):
        self._set_menuitems_status(self._indicator_hdl.get_progress_menu_label, checkpoint)

    def _connect_dbus_signal_handlers(self):
        """Binds DBus signals to their corresponding handler methods.        
        """
        self._indicator_hdl.connect_dbus_event_signal(self.dbus_event_signal_hdl)
        self._indicator_hdl.connect_dbus_error_signal(self.dbus_error_signal_hdl)
        self._indicator_hdl.connect_dbus_progress_signal(self.dbus_progress_signal_hdl)
        self._indicator_hdl.connect_dbus_targetnotfound_signal(self.dbus_targetnotfound_signal_hdl)
        self._indicator_hdl.connect_dbus_alreadyrunning_signal(self.dbus_alreadyrunning_signal_hdl)
        self._indicator_hdl.connect_dbus_exit_signal(self.dbus_exit_signal_hdl)

    def _dbus_check(self):
        _res = True
#        if self._exit is False:
        self._indicator_hdl.test_dbus_validity()
        if not self._indicator_hdl.is_dbus_valid():
            self._notify_info(profilename = "", message = _("Connection to D-Bus service lost."))

            self._set_menuitems_status_unknown()
            self._menuitems["cancel"].set_sensitive(False)

            self._indicator_hdl.dbus_reconnect()
            self._indicator_hdl.test_dbus_validity()
            if self._indicator_hdl.is_dbus_valid():    # re-connection successful?                    
                self._connect_dbus_signal_handlers()
#TODO: get new values here!
            else:
                self._init_dbus_reconnect_timer()
                _res = False
        return _res

    def _dbus_reconnect(self):
        _res = True
#        if self._exit is False:    # do not re-connect when quit was already called
        self._indicator_hdl.test_dbus_validity()
        if self._indicator_hdl.is_dbus_valid():
            _res = False    # already re-connected
        else:
            self._indicator_hdl.dbus_reconnect()
            self._indicator_hdl.test_dbus_validity()
            if self._indicator_hdl.is_dbus_valid():
                self.logger.info(_("re-connection was successful"))
                self._connect_dbus_signal_handlers()
                # get new values here
                self._init_dbus_check_timer()
                _res = False
        return _res

    def _autoexitcheck(self):
        """
        Conditions for autoexit:
        * no dbus avail
        * no backup in progress
        * timeout
        """
        assert self._indicator_hdl.get_keep_alive() is False
        _res = True
        self._indicator_hdl.test_dbus_validity()
        if self._indicator_hdl.is_dbus_valid() is False:
            _running = system.proc_exists(processname = constants.BACKUP_COMMAND)
            if _running is False:
                _res = False
                self._init_autoexit_timer()
        return _res

    def _autoexit(self):
        """
        Conditions for autoexit:
        * no dbus avail
        * no backup in progress
        * timeout
        """
        assert self._indicator_hdl.get_keep_alive() is False
        _res = False
        self._indicator_hdl.test_dbus_validity()
        _running = system.proc_exists(processname = constants.BACKUP_COMMAND)
        if (self._indicator_hdl.is_dbus_valid() is False) and (_running is False):
            if (not self._indicator_hdl.is_error_present()) and\
               (not self._indicator_hdl.is_warning_present()):
                self._notify_info(profilename = "",
                                  message = _("No backup in progress and connection to D-Bus service lost. Simple Backup Indicator is being terminated."))
                gobject.idle_add(self._on_exit)
            else:
                _res = True
        else:
            self._init_autoexitcheck_timer()
        return _res

    def _set_cancel_sensitive(self, sensitive):
        self._menuitems["cancel"].set_sensitive(sensitive)

    def _show_showdialogs_menuitem(self):
        if len(self._current_dialogs) > 0:
            self._menuitems["show_windows"].show()
            self._menuitems["show_windows"].set_sensitive(True)
        else:
            self._menuitems["show_windows"].hide()
            self._menuitems["show_windows"].set_sensitive(False)

    def dbus_event_signal_hdl(self, event, urgency):
        """Method which handles event signals over D-Bus. 
        
        :todo: Implement methods for setting 'finished status' etc. (set menu, disable cancel, hide show
               messages, set icon...
        """
        self._exit = False # eventually interrupt termination
        msg = ""

        if event == 'prepare':
            msg = _("Starting backup session")
            self.set_status_to_normal()
            self._set_menuitems_status_prepare()
            self._set_cancel_sensitive(sensitive = True)

        elif event == 'start':
            self._set_cancel_sensitive(sensitive = True)
            self._indicator_hdl.backup_started()
            self._set_menuitems_status_progress(checkpoint = 0)

        elif event == 'commit':
            pass

        elif event == 'finish':
            msg = _("Ending Backup Session")

            if self._cancel_dialog is not None:
                self._cancel_dialog.destroy()
            self._set_menuitems_status_finished()
            self._set_cancel_sensitive(sensitive = False)
            self.set_status_to_finished()

        elif event == 'backup-canceled':
            self._set_menuitems_status_canceled()
            self._set_cancel_sensitive(sensitive = False)

        else:
            self.logger.warning(_("Unknown D-Bus event `%s` received.") % (event))

        if msg != "":
            profile = self._indicator_hdl.get_profilename()
            self._notify(urgency, profile, msg)

    def set_status_to_normal(self):
        raise NotImplementedError

    def set_status_to_attention(self):
        raise NotImplementedError

    def set_status_to_finished(self):
        raise NotImplementedError

    def _add_dialog_to_showlist(self, dialog):
        if not isinstance(dialog, gtk.Window):
            raise TypeError("GTK window expected.")
        if dialog not in self._current_dialogs:
            self._current_dialogs.append(dialog)
        self._show_showdialogs_menuitem()

    def _remove_dialog_from_showlist(self, dialog):
        if not isinstance(dialog, gtk.Window):
            raise TypeError("GTK window expected.")
        if dialog in self._current_dialogs:
            self._current_dialogs.remove(dialog)
        self._show_showdialogs_menuitem()

    def dbus_targetnotfound_signal_hdl(self):
        self._exit = False # eventually interrupt termination
        self.set_status_to_attention()
        msg = self._indicator_hdl.prepare_targetnotfound_handling()

        assert self._targetnotfound_dialog is None
        self._targetnotfound_dialog = misc.msgdialog_standalone(message_str = "",
                                                                 msgtype = gtk.MESSAGE_ERROR,
                                                                 boxtitle = Infos.NAME,
                                                                 buttons = gtk.BUTTONS_NONE,
                                                                 sticky = True)
        self._targetnotfound_dialog.add_buttons(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                                 _("Try again"), gtk.RESPONSE_OK)

        _ttimer = gobject.timeout_add_seconds(constants.TIMEOUT_RETRY_TARGET_CHECK_SECONDS,
                                    self._targetnotfound_dialog_destroy)
        _ctimer = gobject.timeout_add_seconds(constants.ONE_SECOND,
                                    self._targetnotfound_timer, msg,
                                    constants.ONE_SECOND)

        msg = msg % self._indicator_hdl.get_targetnotfound_clock()

        self._targetnotfound_dialog.set_markup(msg)
        self._add_dialog_to_showlist(dialog = self._targetnotfound_dialog)
        result = self._targetnotfound_dialog.run()
        self._targetnotfound_dialog.destroy()
        self._remove_dialog_from_showlist(dialog = self._targetnotfound_dialog)
        self._targetnotfound_dialog = None

        if result == gtk.RESPONSE_OK:
            retry = constants.RETRY_TRUE
        elif result == gtk.RESPONSE_NONE:
            retry = constants.RETRY_TRUE
        else:
            retry = constants.RETRY_FALSE

        gobject.source_remove(_ttimer)
        gobject.source_remove(_ctimer)
        self._indicator_hdl.finish_targetnotfound_handling(retry)
        self.set_status_to_normal()

    def _targetnotfound_timer(self, msg, interval):
        self._indicator_hdl.decrease_targetnotfound_clock(interval)
        msg = msg % self._indicator_hdl.get_targetnotfound_clock()
        if self._targetnotfound_dialog is not None:
            self._targetnotfound_dialog.set_markup(msg)
        return self._indicator_hdl.get_targetnotfound_run_timer()

    def _targetnotfound_dialog_destroy(self):
        self._indicator_hdl.disable_targetnotfound_run_timer()
        if self._targetnotfound_dialog is not None:
            self._targetnotfound_dialog.destroy()
        return False

    def _notify(self, urgency, profile, message):
        if urgency == 'info':
            self._notify_info(profile, message)

        elif urgency == 'warning':
            self._notify_info(profile, message)
#            self.__indicator_hdl.set_warning_present(is_present = True)
#            self._notify_warning(profile, message)
        else:
            raise ValueError("Unknown urgency!")

    def dbus_error_signal_hdl(self, error):
        self._exit = False # eventually interrupt termination
        self._set_cancel_sensitive(sensitive = False)
        self.set_status_to_attention()
        msg = self._indicator_hdl.get_error_msg(error)
        self._show_errormsg(headline = msg[0], message = msg[1])
        self._indicator_hdl.set_error_present(is_present = False)
        self.set_status_to_normal()
        self._set_menuitems_status_canceled()

    def dbus_progress_signal_hdl(self, checkpoint):
        self._exit = False # eventually interrupt termination
        self._set_menuitems_status_progress(checkpoint)

    def dbus_alreadyrunning_signal_hdl(self):
        self._exit = False # eventually interrupt termination 
        self._notify_info(profilename = "",
                          message = _("Attempt of starting another instance of Simple Backup while this one is already running."))

    def on_showdialogs_clicked(self, *args): #IGNORE:W0613
        self.logger.debug("Show message windows")
        for dialog in self._current_dialogs:
            try:
                dialog.set_keep_above(True)
                dialog.present()
                dialog.set_keep_above(False)
            except AttributeError, error:
                self.logger.warning(_("Unable to present window: %s") % error)

    def on_about_clicked(self, *args): #IGNORE:W0613
        misc.show_about_dialog()

    def on_cancel_clicked(self, *args): #IGNORE:W0613
        """
        :note: We call another method indirectly from here in order NOT to block further
               handling of GUI events (namely 'click on show windows').
        """
        self._set_cancel_sensitive(sensitive = False)
        assert self._cancel_dialog is None
        gobject.idle_add(self._cancel_handler)

    def _cancel_handler(self):
        self._cancel_dialog = misc.msgdialog_standalone(message_str = \
                                    _("Do you really want to cancel the backup process of profile '%s'?") % \
                                    self._indicator_hdl.get_profilename(),
                                    msgtype = gtk.MESSAGE_QUESTION, boxtitle = Infos.NAME,
                                    buttons = gtk.BUTTONS_YES_NO,
                                    headline_str = _("Cancel running backup process?"))
        self._add_dialog_to_showlist(self._cancel_dialog)
        result = self._cancel_dialog.run()
        self._cancel_dialog.destroy()
        self._remove_dialog_from_showlist(self._cancel_dialog)
        self._cancel_dialog = None

        res = True
        if result == gtk.RESPONSE_YES:
            res = self._indicator_hdl.cancel_backup()
            self._set_cancel_sensitive(sensitive = False)
        else:
            self._set_cancel_sensitive(sensitive = True)

        if res is False:
            self._set_cancel_sensitive(sensitive = True)
            self._notify_info(profilename = self._indicator_hdl.get_profilename(),
                              message = _("Unable to cancel backup process."))

    def _show_errormsg(self, headline, message):
        errmsg = misc.errdialog_standalone(message_str = message,
                                          boxtitle = Infos.NAME,
                                          headline_str = headline,
                                          secmsg_str = "")
        self._add_dialog_to_showlist(dialog = errmsg)
        errmsg.run()
        errmsg.destroy()
        self._remove_dialog_from_showlist(dialog = errmsg)
        return False

    def dbus_exit_signal_hdl(self):
        """Handler for D-Bus exit signal.
        """
        self._on_exit()

    def _on_exit(self):
        """Internal method that is called on any exit request. Do not call `__terminate` directly.
        """
        if self._indicator_hdl.get_keep_alive() is True:
            self.logger.debug("Termination of indicator was requested though it is kept alive")
        else:
            self._exit = True
            if (not self._indicator_hdl.is_error_present()) and\
               (not self._indicator_hdl.is_warning_present()):
                gobject.timeout_add_seconds(constants.TIMEOUT_INDICATOR_QUIT_SECONDS, self._terminate)
            else:
                self.logger.debug("still errors present, not calling `__terminate`")

    def _terminate(self):
        """Internal method that actually quits the indicator.
        """
        if self._exit is True:
            self._mainloop.quit()
        return False

    def main(self):
        """The main method.
        """
        self._mainloop.run()
        self._menu.destroy()


class SBackupdIndicatorHandler(object):
    def __init__(self, backup_dbus_obj, options):
        if not isinstance(backup_dbus_obj, dbus_support.DBusClientFacade):
            raise TypeError("Given backup_dbus_obj of type `DBusClientFacade` expected.")

        self.logger = LogFactory.getLogger()
        self.__options = options

        self._backup_dbus_obj = backup_dbus_obj

        self._space_required = constants.SPACE_REQUIRED_UNKNOWN
        self._target = constants.TARGET_UNKNOWN
        self._profile = constants.PROFILE_UNKNOWN

        self.__error_present = False
        self.__warning_present = False

        self._menuitem_status_tmpl = {"profile"        : _("Profile: %s"),
                                      "size_of_backup" : _("Size of backup: %s"),
                                      "progress"       : _("%.1f%% processed"),
                                      "remaining_time" : _("Remaining time: %s")}

        self._targetnotfound_run_timer = False
        self._targetnotfound_clock = 0
        self._starttime_backup = None
        self._time_est_total = 0

    def get_keep_alive(self):
        return self.__options.keep_alive

    def test_dbus_validity(self):
        self._backup_dbus_obj.test_validity()

    def is_dbus_valid(self):
        _valid = self._backup_dbus_obj.get_is_connected()
        return _valid

    def dbus_reconnect(self):
        self._backup_dbus_obj.connect(silent = True)

    def connect_dbus_event_signal(self, handler):
        self._backup_dbus_obj.connect_to_event_signal(handler)

    def connect_dbus_error_signal(self, handler):
        self._backup_dbus_obj.connect_to_error_signal(handler)

    def connect_dbus_progress_signal(self, handler):
        self._backup_dbus_obj.connect_to_progress_signal(handler)

    def connect_dbus_targetnotfound_signal(self, handler):
        self._backup_dbus_obj.connect_to_targetnotfound_signal(handler)

    def connect_dbus_alreadyrunning_signal(self, handler):
        self._backup_dbus_obj.connect_to_alreadyrunning_signal(handler)

    def connect_dbus_exit_signal(self, handler):
        self._backup_dbus_obj.connect_to_exit_signal(handler)

    def __update_properties(self):
        self._profile = self._backup_dbus_obj.get_profilename()
        self._target = self._backup_dbus_obj.get_target()
        self._space_required = self._backup_dbus_obj.get_space_required()

    def cancel_backup(self):
        res = True
        pid = self._backup_dbus_obj.get_backup_pid()
        self.logger.info(_("PID of backup process being canceled: %s") % pid)
        if pid != constants.PID_UNKNOWN:
            term_script = util.get_resource_file(constants.TERMINATE_FILE)
            _ret = system.exec_command_returncode([term_script, str(pid)])
            if _ret != 0:
                self.logger.info(_("Unable to send signal to process %s") % pid)
                self.logger.info(_("Sending signal using gksu to process %s") % pid)
                descr = _("Cancel Backup")
                _ret = system.exec_command_returncode(["gksu", "--description", descr, term_script, str(pid)])
                if _ret != 0:
                    self.logger.warning(_("Sending signal using gksu to process %s failed") % pid)
                    res = False
        else:
            res = False
        return res

    def set_warning_present(self, is_present):
        self.__warning_present = is_present

    def set_error_present(self, is_present):
        self.__error_present = is_present

    def is_error_present(self):
        return self.__error_present

    def is_warning_present(self):
        return self.__warning_present

    def __set_retry_target_check(self, retry):
        if retry not in (constants.RETRY_UNKNOWN, constants.RETRY_FALSE,
                         constants.RETRY_TRUE):
            raise ValueError("Invalid value for retry target check.")
        self._backup_dbus_obj.set_retry_target_check(retry)

    def enable_targetnotfound_run_timer(self):
        self._targetnotfound_run_timer = True

    def disable_targetnotfound_run_timer(self):
        self._targetnotfound_run_timer = False

    def get_targetnotfound_run_timer(self):
        return self._targetnotfound_run_timer

    def get_targetnotfound_error_msg(self):
        target = self._backup_dbus_obj.get_target()
        msg = _("<b>Unable to find specified target directory</b>\n\nThe specified target directory '%s' was not found.\n\n") % (glib.markup_escape_text(target))
        msg = msg + _("You can try to use the specified target again or cancel the profile execution. The specified destination is automatically used in %s seconds.")
        return msg

    def prepare_targetnotfound_handling(self):
        self.__update_properties()
        self.set_error_present(is_present = True)
        self.__set_retry_target_check(constants.RETRY_UNKNOWN)
        msg = self.get_targetnotfound_error_msg()

        self.enable_targetnotfound_run_timer()

        self._targetnotfound_clock = constants.TIMEOUT_RETRY_TARGET_CHECK_SECONDS
        return msg

    def finish_targetnotfound_handling(self, retry):
        self.__set_retry_target_check(retry)
        self.set_error_present(is_present = False)
        self.disable_targetnotfound_run_timer()

    def decrease_targetnotfound_clock(self, interval):
        self._targetnotfound_clock -= interval

    def get_targetnotfound_clock(self):
        return self._targetnotfound_clock

    def get_profilename(self):
        if self._profile == constants.PROFILE_UNKNOWN:
            _res = _("unknown")
        else:
            _res = self._profile
        return _res

    def get_progress_menu_label(self, checkpoint):
        self.__update_properties()
        checkpoint = int(checkpoint)
        space_str = self.__get_space_required_str()

        # values valid?
        if (self._space_required > 0) and (checkpoint > 0):
            _done = (checkpoint - 1) * constants.TAR_RECORDSIZE
            _percent = (float(_done) / float(self._space_required)) * 100.0
            _percent = min(100.0, max(0.01, _percent))

            _time_passed = time.time() - self._starttime_backup
            _time_est_total = (_time_passed / _percent) * 100.0
            _time_est_remain = _time_est_total - _time_passed

            if _time_est_remain < 60:
                _time_str = _("less than 1 minute")
            elif _time_est_remain < 120:
                _time_str = _("about 2 minutes")
            elif _time_est_remain < 180:
                _time_str = _("about 3 minutes")
            elif _time_est_remain < 240:
                _time_str = _("about 4 minutes")
            elif _time_est_remain < 300:
                _time_str = _("about 5 minutes")
            else:
                _time_str = _("about %.0f minutes") % (round((_time_est_remain / 60)))

            menu_msg = {"profile"        : self._menuitem_status_tmpl["profile"] % self.get_profilename(),
                        "size_of_backup" : self._menuitem_status_tmpl["size_of_backup"] % space_str,
                        "progress"       : self._menuitem_status_tmpl["progress"] % _percent,
                        "remaining_time" : self._menuitem_status_tmpl["remaining_time"] % _time_str}

        else: # values invalid, but something is in progress
            menu_msg = {"profile"        : self._menuitem_status_tmpl["profile"] % self.get_profilename(),
                        "size_of_backup" : self._menuitem_status_tmpl["size_of_backup"] % space_str,
                        "progress"       : _("In progress"),
                        "remaining_time" : self._menuitem_status_tmpl["remaining_time"] % _("unknown")}
        return menu_msg

    def get_finished_menu_label(self):
        self.__update_properties()
        space_str = self.__get_space_required_str()
        menu_msg = {"profile"        : self._menuitem_status_tmpl["profile"] % self.get_profilename(),
                    "size_of_backup" : self._menuitem_status_tmpl["size_of_backup"] % space_str,
                    "progress"       : self._menuitem_status_tmpl["progress"] % 100,
                    "remaining_time" : self._menuitem_status_tmpl["remaining_time"] % _("finished")}
        return menu_msg

    def get_unknown_menu_label(self):
        self.__update_properties()
        space_str = self.__get_space_required_str()
        menu_msg = {"profile"        : self._menuitem_status_tmpl["profile"] % self.get_profilename(),
                    "size_of_backup" : self._menuitem_status_tmpl["size_of_backup"] % space_str,
                    "progress"       : _("Progress: unknown"),
                    "remaining_time" : self._menuitem_status_tmpl["remaining_time"] % _("unknown")}
        return menu_msg

    def get_prepare_menu_label(self):
        self.__update_properties()
        menu_msg = {"profile"        : self._menuitem_status_tmpl["profile"] % self.get_profilename(),
                    "size_of_backup" : _("Backup is being prepared"),
                    "progress"       : _("Collecting file info"),
                    "remaining_time" : self._menuitem_status_tmpl["remaining_time"] % _("unknown")}
        return menu_msg

    def get_canceled_menu_label(self):
        self.__update_properties()
        space_str = self.__get_space_required_str()
        menu_msg = {"profile"        : self._menuitem_status_tmpl["profile"] % self.get_profilename(),
                    "size_of_backup" : self._menuitem_status_tmpl["size_of_backup"] % space_str,
                    "progress"       : _("Progress: canceled"),
                    "remaining_time" : self._menuitem_status_tmpl["remaining_time"] % _("finished")}
        return menu_msg

    def __get_space_required_str(self):
        if self._space_required > 0:
            space_str = util.get_humanreadable_size_str(size_in_bytes = self._space_required,
                                                        binary_prefixes = True)
        else:
            space_str = _("unknown")
        return space_str

    def get_error_msg(self, error):
        """Returns tuple of error message:
        1. Headline
        2. actual error message
        3. additional info
        
        :todo: rename into `format_error_msg`
        """
        self.__update_properties()
        self.set_error_present(is_present = True)
        msg = (_("An error occurred"), error, "")
        return msg

    def backup_started(self):
        self.__update_properties()
        self._starttime_backup = time.time()
        self._time_est_total = 0


class SBackupdIndicatorApp(object):
    """GUI for listen to the backup daemon. It uses DBus service. 
    """
    def __init__(self, options):
        self.__options = options
        self.__indicator_gui = None
        self.__indicator_hdl = None
        self.__dbus = dbus_support.DBusClientFacade(constants.INDICATORAPP_NAME)
        self.__lock = lock.ApplicationLock(lockfile = constants.LOCKFILE_INDICATOR_FULL_PATH,
                                           processname = constants.INDICATORAPP_FILE, pid = os.getpid())

    def main(self, indicator_class):
        exitcode = constants.EXCODE_SUCCESS
        try:
            self.__lock.lock()
            self.__dbus.connect()    # establish D-Bus connection

            self.__indicator_hdl = SBackupdIndicatorHandler(self.__dbus, self.__options)
            self.__indicator_gui = indicator_class(self.__indicator_hdl)
            self.__indicator_gui.main()


        except exceptions.InstanceRunningError:
            exitcode = constants.EXCODE_INSTANCE_ALREADY_RUNNING
            print _("Another `Simple Backup Indicator` is already running.")

        except SystemExit, error:
            exitcode = error.code

        except KeyboardInterrupt:
            exitcode = constants.EXCODE_KEYBOARD_INTERRUPT

        finally:
            self.__dbus.quit()   # we connect here and we quit here
            self.__lock.unlock()

        return exitcode


def main(options, indicator_class):
    exitcode = constants.EXCODE_GENERAL_ERROR

    dbus.mainloop.glib.DBusGMainLoop(set_as_default = True)
    gtk_avail = True
    try:
        gtk.init_check()
    except RuntimeError, error:
        gtk_avail = False
        print _("Initialization of GTK+ failed: %s") % error

    if gtk_avail:
        sbdgui = SBackupdIndicatorApp(options)
        exitcode = sbdgui.main(indicator_class)

    return exitcode



#   Simple Backup - general observer-notfifier implementation
#
#   Copyright (c)2009-2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


from sbackup.util import constants


# Handling of states and information provided by Dbus objects is
# not ideal. Improve!


class Subject(object):
    def __init__(self):
        self._observers = []

    def attach(self, observer):
        if not observer in self._observers:
            self._observers.append(observer)

    def detach(self, observer):
        if observer in self._observers:
            self._observers.remove(observer)

    def notify(self, modifier = None):
        for observer in self._observers:
            if modifier != observer:
                observer.update(self)



class Observer(object):
    def update(self, subject):
        raise NotImplementedError("Abstract method must be overridden in "\
                                  "derived class!")


class SBackupState(Subject):
    """Represents a single state and according data within
    the backup process. Whenever interesting data changes
    one can change the state and it is published to registered
    observers.
    
    :todo: Add type checking in setter methods!
    
    """
    def __init__(self):
        Subject.__init__(self)

        self.__state = 'unknown'
        self.__urgency = 'info'
        self.__profilename = constants.PROFILE_UNKNOWN
        self.__space_required = constants.SPACE_REQUIRED_UNKNOWN
        self.__is_full = -1
        self.__target = constants.TARGET_UNKNOWN
        self.__recent_error = 'None'
        self.clear_backup_properties()

    def clear_backup_properties(self):
        self.__profilename = constants.PROFILE_UNKNOWN
        self.__space_required = constants.SPACE_REQUIRED_UNKNOWN
        self.__is_full = -1
        self.__target = constants.TARGET_UNKNOWN

    def set_state(self, state):
        """Observers are only notified, if the state has changed.
        
        """
        self.__state = state
        self.notify()

    def get_state(self):
        return self.__state

    def set_urgency(self, urgency):
        """The urgency is used to distinguish several types of events.
        There are following urgencies:
        
        -critical
        -error
        -warning
        -info.
        
        """
        self.__urgency = urgency

    def get_urgency(self):
        return self.__urgency

    def set_profilename(self, profilename):
        self.__profilename = profilename

    def get_profilename(self):
        return self.__profilename

    def set_recent_error(self, recent_error):
        self.__recent_error = recent_error

    def get_recent_error(self):
        return self.__recent_error

    def set_is_full(self, is_full):
        self.__is_full = is_full

    def set_target(self, target):
        self.__target = target

    def get_target(self):
        return self.__target

    def set_space_required(self, space):
        self.__space_required = space

    def get_space_required(self):
        return self.__space_required



#   Simple Backup - implementation of threaded worker tasks
#
#   Copyright (c)2009-2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

"""This module contains definitions of classes helping to process
worker tasks within separate threads.
"""

import threading

from sbackup.util import log


class WorkerThread(object):
    """Class that encapsulates a single (long running) thread.
    The current implementation does not support control of the task's
    progress (due to it isn't necessary for tar operations).
    """

    def __init__(self, task):
        """Default constructor.

        @param task: a callable object that actually represents the worker task
        @type task:  callable object
        
        @raise TypeError: in case of type mismatch of given parameters
        """
        self.__task = None
        self.__finish_callback = None
        self.__finish_args = None
        self.__thread = None
        self.__set_task(task)

    def __set_task(self, callback):
        """Private helper method that sets the callback function representing
        the actual worker task.
        
        @param callback: callable object representing the task to perform
        @type callback:  callable object
        
        @return: None
        
        @raise TypeError: If the given parameter is not callable
        """
        if not callable(callback):
            raise TypeError("The given task object object is not callable!")
        self.__task = callback

    def set_finish_callback(self, callback, *args):
        """Sets the given object as callback function/method that is
        invoked after finishing the worker thread. An arbitrary number of
        non-keyword parameters can be given. Since the worker task is
        performed in a separate thread one need to set a callback function
        e.g. for evaluating any results of the task.
        
        The result of the worker task is given as last parameter
        to this callback function, this is even valid for tasks that
        return 'nothing', i.e. None.

        @param callback: callable object that is called after finishing
        @type callback: callable object
        @param args: arbritrary number of parameters given to the callback
                     function
        
        @return: None

        @raise TypeError: If the given callback function is not callable
        """
        if not callable(callback):
            raise TypeError("The given callback object is not callable!")
        self.__finish_callback = callback
        self.__finish_args = args

    def __run_task(self, *args, **kwargs):
        """Private method that executes the worker task and afterwards
        calls the specified finish callback function (if any).
        
        Any exception raised within the worker task is catched and returned
        as result of the task. If no exception occurs the result of the
        worker task is returned as result. To determine if an exception was
        raised within the thread, check the type of the returned result.

        The given keyword and non-keyword parameters are given (as they are)
        to the worker task.
        
        @param args:   non-keyword parameters given to the worker task
        @param kwargs: keyword parameters given to the worker task
        
        @return: None        
        """
        try:
            result = self.__task(*args, **kwargs)
        except Exception, exc:
            # if an exception was raised, use it as the result
            log.LogFactory.getLogger().exception(exc)
            result = exc

        final_args_lst = []
        # if there is a function to call after finishing?
        if self.__finish_callback is not None:
            if self.__finish_args is not None:
                final_args_lst = [ _arg for _arg in self.__finish_args ]
            # append the worker task result to the arguments list
            final_args_lst.append(result)

            # now the final arguments list is converted into a tuple
            final_args = tuple(final_args_lst)
            self.__finish_callback(*final_args)

    def start(self, *args, **kwargs):
        """Public method that starts this worker thread. The given keyword
        and non-keyword parameters are given (as they are) to the worker task.

        @param args:   non-keyword parameters given to the worker task
        @param kwargs: keyword parameters given to the worker task
        
        @return: None
        """
        self.__thread = threading.Thread(target = self.__run_task,
                                          args = args, kwargs = kwargs)
        self.__thread.start()



#   Simple Backup - unified file handling using (Python's) native functions
#
#   Copyright (c)2009-2010,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2009: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`local_file_utils` --- unified file handling (local)
====================================================================

.. module:: local_file_utils
   :synopsis: common functions for unified file handling within local filesystems
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>
.. moduleauthor:: Ouattara Oumar Aziz (alias wattazoum) <wattazoum@gmail.com>

"""


import os.path
import shutil
import pickle
import stat
import types

import gzip
import uuid


from sbackup.util import pathparse
from sbackup.util import exceptions
from sbackup.util import constants
from sbackup.util import system
from sbackup.util import log



#TODO: make these functions module functions? See article: Python is not Java

PATHSEP = system.PATHSEP

path_exists = os.path.exists
is_dir = os.path.isdir
listdir = os.listdir
makedir = os.mkdir #, 0750)
makedirs = os.makedirs #, 0750)
chmod = os.chmod
get_dirname = os.path.dirname
get_basename = os.path.basename
is_mount = os.path.ismount

# TDOD: Evaluate alternate implementations:
#def path_exists(path):
# os.path.exists(path)
# os.access(uri, os.F_OK)

#def copyfile(src, dest):
#    """Copies given file and metadata
#    """
#    if os.path.isdir(src):
#        raise ValueError("Given copy source is a directory, not a file")
##TODO: What happens if `dest` already exists?
#    if os.path.isfile(src):
#        shutil.copy2(src, dest)

def copyfile(src, dst):
    """Copies given file and metadata including file permissions.
    If this fails a custom exception is raised. 
    
    @param src: an existing file that should be copied
    @param dst: copy destination - an existing directory or full path to new file
                
    @return: None
    
    @raise CopyFileAttributesError: if the permissions could not be copied
    """
    # only absolute paths are supported
    if not os.path.isabs(src):
        raise ValueError("Given copy source `%s` must be absolute" % src)
    if not os.path.isabs(dst):
        raise ValueError("Given copy destination `%s` must be absolute" % dst)

    # the source must be a file and exist
    if not os.path.exists(src):
        raise IOError("Given copy source `%s` does not exist" % src)
    if os.path.isfile(src):
        prep_src, prep_dst = _prepare_copy(src, dst)
        shutil.copyfile(prep_src, prep_dst)
        try:
            shutil.copystat(prep_src, prep_dst)
        except OSError:
            raise exceptions.CopyFileAttributesError(\
                            "Unable to copy file attributes (permissions etc.) of file `%s`." % src)
    else:
        _logger = log.LogFactory.getLogger()
        _logger.warning("Given copy source `%s` is not a file. Skipped." % src)


def _prepare_copy(src, dst):
    """Helper function that prepares the given paths for copying
    using 'nssb_copy'.
    
    Source must be a file or symbolic link to a file!
    
    @todo: Implement test case for symbolic links!
    """
    _src_file = os.path.basename(src)
    _src_dir = os.path.dirname(src)

    if os.path.isdir(dst):
        _dst_file = _src_file
        _dst_dir = dst
    elif dst.endswith(os.path.sep):
        _dst_file = _src_file
        _dst_dir = dst
    else:
        _dst_file = os.path.basename(dst)
        _dst_dir = os.path.dirname(dst)

    if not os.path.isdir(_dst_dir):
        raise IOError("Given copy destination '%s' does not exist" % _dst_dir)

    _dst_path = joinpath(_dst_dir, _dst_file)
    retval = (src, _dst_path)

    return retval

def listdir_fullpath(path) :
    """List a directory. Returns full paths to entries.
    """
    _lst = listdir(path)
    _res = []
    for _ent in _lst:
        _res.append(joinpath(path, _ent))
    return _res


def path_writeable(path):
    _res = os.access(path, os.W_OK)
    return _res

def normpath(*args):
    """Normalizes the given paths (i.e. concatenates them and removes trailing
    separators).
    
    @todo: Consistent handling of `normpath` (Quote: os.path.normpath - It should be understood
    that this may change the meaning of the path if it contains symbolic links!)
    
    :note: Be careful when using `normpath`. Consider `joinpath` instead.
    """
    _path = os.path.join(*args)
    _path = pathparse.remove_trailing_sep(_path)
#    _path = os.path.normpath(_path)
    return _path

def joinpath(*args):
    return pathparse.joinpath(*args)

def is_link(path):
    spath = pathparse.remove_trailing_sep(path)
    res = os.path.islink(spath)
    return res

def get_link(path):
    """Returns the target of given link `path`. Relative links remain
    unchanged (i.e. are not made absolute). 
    """
    if not is_link(path):
        raise ValueError("Given path is not a symbolic link.")
    _res = os.readlink(path)
    return _res

def get_link_abs(path):
    """Returns the absolute target of given link `path`. Relative links are
    modified (i.e. are made absolute). 
    """
    _ln_target = get_link(path)
    if os.path.isabs(_ln_target):
        _res = _ln_target
    else:
        _res = os.path.join(os.path.dirname(path), _ln_target)
    _res = os.path.abspath(_res)
    return _res

def delete(uri):
    """Deletes given file or directory (recursive).
    """
    if os.path.isdir(uri) and not os.path.islink(uri):
        shutil.rmtree(uri, ignore_errors = False)    #, onerror=_onerror)
    else:
        os.unlink(uri)

def force_delete(path):
    """Forces removal of given file or directory (recursive).
    """
    _add_write_permission(path, recursive = True)
    delete(path)

def destinsrc(src, dst):
    src = os.path.abspath(src)
    dst = os.path.abspath(dst)
    if not src.endswith(os.path.sep):
        src += os.path.sep
    if not dst.endswith(os.path.sep):
        dst += os.path.sep
    return dst.startswith(src)

def force_move(src, dst):
    """Modified version of `shutil.move` that uses `nssb_copytree`
    and even removes read-only files/directories.
    :note: it does not work (and won't never) if the `src` is located within a read-only
            directory. We'd need to manipulate the parent dir in that case.
    """
    try:
        os.rename(src, dst)
    except OSError:
        if os.path.isdir(src):
            if destinsrc(src, dst):
                raise shutil.Error("Cannot move a directory '%s' into itself "\
                                   "'%s'." % (src, dst))
#            _copytree(src, dst, symlinks = True)
#FIXME: copy symlinks (not the target)
            _copytree(src, dst, symlinks = False)
            force_delete(src)
        else:
            shutil.copy2(src, dst)
            force_delete(src)

def _copytree(src, dst, symlinks = False):
    """mod of `shutil.copytree`. This doesn't fail if the
    directory exists, it copies inside.

    :param src: source path for copy operation
    :param dst: destination
    :param symlinks: copy symlinks?
    :type src: string
    :type dst: string

    """
    names = os.listdir(src)
    if not os.path.exists(dst) :
        os.makedirs(dst)
    errors = []
    for name in names:
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if symlinks and os.path.islink(srcname):
                linkto = os.readlink(srcname)
                os.symlink(linkto, dstname)
            elif os.path.isdir(srcname):
                _copytree(srcname, dstname, symlinks)
            else:
                shutil.copy2(srcname, dstname)
            # XXX What about devices, sockets etc.?
        except (IOError, OSError), why:
            errors.append((srcname, dstname, str(why)))
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except shutil.Error, err:
            errors.extend(err.args[0])
    try:
        shutil.copystat(src, dst)
    except OSError, why:
        errors.extend((src, dst, str(why)))
    if len(errors) > 0:
        raise shutil.Error, errors

def createfile(filepath):
    """
    Create an empty file
    @param filepath : The file path
    
    :todo: Review and improve!
    """
    filepath = filepath.rstrip(os.sep)
    spl = filepath.split(os.sep)
    if len(spl) == 1 :
        writetofile(filepath, "")
    else:
        _dir = filepath[:-len(spl[len(spl) - 1])]
        if path_exists(_dir):
            writetofile(filepath, "")
        else:
            makedir(_dir)
            writetofile(filepath, "")

def readfile(path):
    """Read a file from a given URI and returns a string with the read content.
    
    @rtype: String
    """
    fobj = open(path, "r")
    value = fobj.read()
    fobj.close()
    return value

def writetofile(path, content) :
    """
    Write a String to a file. You don't have to open and close the file.
    - File = path to the file
    - StringToWrite = String to write into the file.
    """
    fobj = open(path, "w")
    fobj.write(content)
    fobj.close()

def openfile(uri, write = False):
    " opens a file for reading or writing. Default is reading "
    if write:
        return open(uri, "w")
    else:
        return open(uri, "r")

def openfile_for_append(path):
    return open(path, "a+")

def pickleload(path):
    """
    Load a python object from the given pickle file
    @param file: the path of the pickle file
    """
    f = openfile(path)
    result = pickle.load(f)
    f.close()
    return result

def pickledump(datas, path):
    """
    Dump the given datas into the file given 
    @param datas: any type of python datas/object
    @param file : a file path to the file in wich the dump will be made
    """
    f = openfile(path, True)
    pickle.dump(datas , f)
    f.close()

def _add_write_permission(path, recursive = True):
    """Sets write permissions for user, group, and others for
    given directory or file (recursive). 
    """
    fstats = os.stat(path)
    fmode = fstats.st_mode
    fmode = fmode | system.UNIX_PERM_ALL_WRITE
    os.chmod(path, fmode)

    if os.path.isdir(path) and recursive is True:
        for _entry in os.listdir(path):
            _entrypath = os.path.join(path, _entry)
            if os.path.isdir(_entrypath)  and not os.path.islink(_entrypath):
                _add_write_permission(_entrypath)
            else:
                _add_write_permission(_entrypath, recursive = False)

def chmod_no_rwx_grp_oth(path):
    """Sets write permissions for user only for
    given directory or file (*not* recursive). 
    """
    try:
        fstats = os.stat(path)
        fmode = fstats.st_mode
        fmode = fmode & system.UNIX_PERM_GRPOTH_NORWX
        os.chmod(path, fmode)
    except Exception, error:
        _msg = "Unable to set permissions: %s" % str(error)
        _logger = log.LogFactory.getLogger()
        _logger.warning(_msg)
        return

def rename(src, dst):
    # avoids (misused) move operations using `rename`
    _dstf = joinpath(get_dirname(src), get_basename(dst))
    os.rename(src, _dstf)

def rename_errors_ignored(src, dst):
    try:
        rename(src, dst)
    except OSError:
        pass

def rename_rotating(src, dst, max_num, compressed = True):
    """Renames the given file `src` to `dst`. The destination (i.e. the new
    file name) is renamed in rotated manner prior the actual renaming
    process.
    If `compressed` is set to True, compressed files (*.gz) are considered.
    """
    if not isinstance(src, types.StringTypes):
        raise TypeError("Expected string as source. Got %s instead." % type(src))
    if not isinstance(dst, types.StringTypes):
        raise TypeError("Expected string as destination. Got %s instead." % type(dst))
    if not isinstance(max_num, types.IntType):
        raise TypeError("Expected integer as max. number. Got %s instead." % type(max_num))
    if max_num < 1:
        raise ValueError("Max. number must be greater than 0.")

    # at first: rotating rename of destination
    for _num in range(max_num, 0, -1):
        _rot_src = pathparse.append_str_to_filename(dst, str((_num - 1)))
        _rot_target = pathparse.append_str_to_filename(dst, str(_num))
        if compressed is True:
            _rot_src = "%s.gz" % _rot_src
            _rot_target = "%s.gz" % _rot_target
        if path_exists(_rot_src):
            rename(_rot_src, _rot_target)
    # then rename the source file
    rename(src, dst)


def compress_rotated_files(basename, max_num):
    """Compresses files with trailing number 0..max_num.
    """
    if not isinstance(basename, types.StringTypes):
        raise TypeError("Expected string as basename. Got %s instead." % type(basename))
    if not isinstance(max_num, types.IntType):
        raise TypeError("Expected integer as max. number. Got %s instead." % type(max_num))
    if max_num < 1:
        raise ValueError("Max. number must be greater than 0.")

    for _num in range(max_num, -1, -1):
        _src = pathparse.append_str_to_filename(basename, str(_num))
        if path_exists(_src):
            compress(_src)


def compress(src, keep_original = False):
    if not isinstance(src, types.StringTypes):
        raise TypeError("Expected string as source. Got %s instead." % type(src))

    out_file = "%s.gz" % src
    if path_exists(src) and os.path.isfile(src):
        f_in = open(src, 'rb')
        f_out = gzip.open(out_file, 'wb')
        f_out.writelines(f_in)
        f_out.close()
        f_in.close()
        if not keep_original:
            delete(src)


def stat_device(path):
    _res = os.stat(path)[stat.ST_DEV]
    return _res

def stat_inode(path):
    _res = os.stat(path)[stat.ST_INO]
    return _res



def test_path(path, testdir_name, testfile_name, test_read = True):
    __logger = log.LogFactory().getLogger()

    testdir = os.path.join(path, testdir_name)
    testfile = os.path.join(testdir, testfile_name)

    __logger.info("Perform tests at specified location")
    try:
        __logger.debug("test access to specified path using native functions")
        _exists = path_exists(path)
        if bool(_exists) is False:
            raise exceptions.RemoteMountTestFailedError("Specified remote path is not accessable.")

        # test directory
        __logger.debug("Test testdir: %s" % testdir)
        _exists = path_exists(testdir)
        if bool(_exists) is True:
            raise exceptions.RemoteMountTestFailedError("Unable to create directory for testing purpose: Directory already exists.")

        __logger.debug("Create testdir")
        makedir(testdir)

        __logger.debug("Test testfile for existence")
        _exists = path_exists(testfile)
        if bool(_exists) is True:
            raise exceptions.RemoteMountTestFailedError("Unable to create file for testing purpose: File already exists.")

        _buffer = "Some arbitrary content: %s" % uuid.uuid4()
        __logger.debug("Write buffer: `%s` to file" % _buffer)
        writetofile(testfile, _buffer)

        if test_read is True:
            # and re-read
            __logger.debug("Re-read test file")
            _exists = path_exists(testfile)
            if bool(_exists) is False:
                raise exceptions.RemoteMountTestFailedError("Unable to open file for testing purpose: File does not exists.")
            _cont = readfile(testfile)
            if _cont != _buffer:
                raise exceptions.RemoteMountTestFailedError("Unable to read content from test file: content differs.")

        # clean-up
        __logger.debug("Remove file")
        delete(testfile)
        __logger.debug("Remove dir")
        delete(testdir)
    except (OSError, IOError), error:
        raise exceptions.RemoteMountTestFailedError(str(error))


def query_fs_info(path):
    _logger = log.LogFactory.getLogger()
    _size = constants.SIZE_FILESYSTEM_UNKNOWN
    _free = constants.FREE_SPACE_UNKNOWN

    try:
        _vstat = os.statvfs(path)
    except OSError, error:
        _logger.error("Error in `query_fs_info`: %s" % error)
    else:
        _size = _vstat.f_blocks * _vstat.f_bsize
        _free = _vstat.f_bavail * _vstat.f_bsize
    _logger.debug("FS info - size: %s free: %s" % (_size, _free))
    return (_size, _free)
